

class FilterType(object):
    dss = 'dss'  # hotelling-t2 test in the time-domain

    def get_available_methods(self):
        members = [attr for attr in dir(self) if not callable(getattr(self, attr)) and not attr.startswith("__")]
        return members


class DSSData(object):
    def __init__(self,
                 filter_type='DSS',
                 bias_frequencies=None,
                 bias_domain=None,
                 unbiased_power=None,
                 biased_power=None,
                 total_unbiased_power=None,
                 total_biased_power=None,
                 n_components_rotation_1=None,
                 n_components_rotation_2=None,
                 n_possible_components=None,
                 threshold=None,
                 component_rank=None,
                 power_ratio=None,
                 kept=True,
                 main_channel=None):
        self.filter_type = filter_type
        self.bias_frequencies = bias_frequencies
        self.bias_domain = bias_domain
        self.unbiased_power = unbiased_power
        self.biased_power = biased_power
        self.total_unbiased_power = total_unbiased_power
        self.total_biased_power = total_biased_power
        self.n_components_rotation_1 = n_components_rotation_1
        self.n_components_rotation_2 = n_components_rotation_2
        self.n_possible_components = n_possible_components
        self.threshold = threshold
        self.component_rank = component_rank
        self.power_ratio = power_ratio
        self.kept = kept
        self.main_channel = main_channel



"""
This example uses an estimation of the impulse response in order to estimate and subtract the stimulus artifact.
Here we use a noise as the stimulus and its envelope plus a delay as the brain response.
"""
import numpy as np
import matplotlib.pyplot as plt
from peegy.processing.tools.epochs_processing_tools import et_regression_subtraction
from peegy.processing.tools.filters.eegFiltering import bandpass_fir_win
from peegy.processing.tools.multiprocessing.multiprocessesing_filter import filt_data

fs = 16384.0
epoch_length = 1.0
epoch_length = np.ceil(epoch_length * fs) / fs  # fit to fs rate
burst_duration = 0.375
ffr_frequencies = np.array([120, 240, 360, 480, 600.0])
ffr_frequencies = np.ceil(burst_duration * ffr_frequencies) / burst_duration  # fit to burst_duration length
alternating_polarity = False  # stimulus is alternating in polarity every presentation
# here we pick some random frequencies to test statistical detection
random_frequencies = np.unique(np.random.randint(100, 400, 3))
stim_delay = 0.001  # neural delay in secs
brain_delay = 0.0237
time = np.arange(0, burst_duration, 1 / fs).reshape(-1, 1)
# stimulation waveform
_original_stimulus_waveform = np.sum(
    3 * np.sin(2 * np.pi * ffr_frequencies * time),
    axis=1).reshape(-1, 1) * 1.0  # generates ffr artifacts with 1 uV amplitude
stimulus_waveform = np.pad(_original_stimulus_waveform, ((0, int(fs * brain_delay) + int(fs * stim_delay)),
                                                         (0, 0)), 'constant', constant_values=(0, 0))

# apply delay to leaked stimulus (assuming system delay)
_leak_stimulus = np.pad(_original_stimulus_waveform, ((int(fs * stim_delay), int(fs * brain_delay)),
                                                      (0, 0)), 'constant', constant_values=(0, 0))
# brain response
template_waveform = np.pad(_original_stimulus_waveform, ((int(fs * stim_delay) + int(fs * brain_delay), 0), (0, 0)),
                           'constant', constant_values=(0, 0))
_b_brain = bandpass_fir_win(high_pass=60, low_pass=800, fs=fs)
template_waveform = filt_data(data=template_waveform, b=_b_brain)
template_waveform *= 0.2  # 0.2 uV amplitude and a delay
template_waveform[template_waveform < 0] = 0  # rectify

# leaked artifact
_b = bandpass_fir_win(0.1, 1000, fs=fs)
_leak_stimulus = filt_data(data=_leak_stimulus, b=_b)

# scale respective signals
template_waveform *= 0.2
_leak_stimulus *= 1.0

# generate recorded data including brain, leaked artifact and noise
snr = 0  # dB
np.random.seed(2)
noise = 10 ** (-snr/20) * np.random.randn(template_waveform.size, 1) * np.std(template_waveform)

recorded_waveform = template_waveform + _leak_stimulus + noise

# # zero padding
stimulus_waveform = np.pad(stimulus_waveform, ((0, 0*stimulus_waveform.shape[0]),
                                               (0, 0)), 'constant', constant_values=(0, 0))
recorded_waveform = np.pad(recorded_waveform, ((0, 0*recorded_waveform.shape[0]),
                                               (0, 0)), 'constant', constant_values=(0, 0))

# plot stimuli, brain and leaked artifact
plt.plot(stimulus_waveform, label='input waveform')
plt.plot(recorded_waveform, label='recorded data')
plt.plot(_leak_stimulus, label='leaked artifact')
plt.plot(template_waveform, label='brain response')
plt.legend()
plt.show()

# remove artifacts
_recovered_response, _recovered_artifact = et_regression_subtraction(
    data=np.atleast_3d(recorded_waveform),
    stimulus_waveform=stimulus_waveform,
    max_lag=50,
    max_length=200,
    plot_results=False)

# apply filter to extract neural bandwidth
_recovered_response = filt_data(data=_recovered_response, b=_b_brain)

# apply filter to extract neural response without removing artifact
_contaminated_response = filt_data(data=recorded_waveform, b=_b_brain)

# apply filter to extract neural response without removing artifact
_filtered_artifact = filt_data(data=_leak_stimulus, b=_b_brain)


# plot artifact
plt.figure()
plt.plot(_leak_stimulus, label='target artifact')
plt.plot(_recovered_artifact.flatten(), label='recovered artifact')
plt.legend()
plt.show()

# plot response
plt.figure()
plt.plot(_contaminated_response, label='contaminated response Corr: {:}'.format(
    np.corrcoef(template_waveform.flatten(),
                _contaminated_response[0:template_waveform.shape[0]].flatten())[-1, 0]))

plt.plot(_filtered_artifact, label='leaked artifact response Corr: {:}'.format(
    np.corrcoef(template_waveform.flatten(),
                _filtered_artifact[0:template_waveform.shape[0]].flatten())[-1, 0]))

plt.plot(template_waveform, label='target response')

plt.plot(_recovered_response.flatten(), label='recovered response. Corr: {:}'.format(
    np.corrcoef(template_waveform.flatten(),
                _recovered_response[0:template_waveform.shape[0]].flatten())[-1, 0]))
plt.legend()
plt.show()



import re
from pathlib import Path
from peegy.definitions.events import Events
from peegy.io.exporters.edf_bdf_writer import data_to_bdf
from peegy.processing.pipe.io import ReadInputData
import pandas as pd
import json
import os
import shutil
import astropy.units as u
from typing import List


class BIDS_EEG_JSON(object):
    def __init__(self,
                 TaskName: str = "Hearing sounds",
                 InstitutionName: str = "Interacoustics",
                 InstitutionAddress: str | None = None,
                 InstitutionalDepartmentName: str | None = None,
                 Manufacturer: str = "Interacoustics",
                 ManufacturersModelName: str = "Eclipse",
                 SoftwareVersions: str | None = None,
                 TaskDescription: str = "Passive experiment, the participants is not required to do anything",
                 Instructions: str = "Be relax and quiet while you hear the sounds",
                 CogAtlasID: str | None = None,
                 CogPOID: str | None = None,
                 DeviceSerialNumber: str | None = None,
                 EEGReference: str = "single electrode placed on FCz",
                 SamplingFrequency: float = 2400,
                 PowerLineFrequency: float = 50,  # number or "n/a"
                 SoftwareFilters: dict | str = "n/a",  # {"Anti-aliasing filter": {"half-amplitude cutoff (Hz)": 500,
                 # "Roll-off": "6dB/Octave"}},  # object of objects or "n/a"
                 CapManufacturer: str = "None",
                 CapManufacturersModelName: str = "None",
                 EEGChannelCount: int = 1,
                 ECGChannelCount: int = 0,
                 EMGChannelCount: int = 0,
                 EOGChannelCount: int = 0,
                 MiscChannelCount: int = 0,
                 TriggerChannelCount: int = 1,
                 RecordingDuration: float | None = None,
                 RecordingType: str = "continuous",
                 EpochLength: float | None = None,
                 EEGGround: str = "placed on participant's cheek",
                 HeadCircumference: float | None = None,
                 EEGPlacementScheme: str = "10-20 system",
                 HardwareFilters: dict | str = "n/a",  # {"ADC's decimation filter (hardware bandwidth limit)":{": str -
                 # 3dB cutoff point (Hz)":480, ": str F ilter order sinc response":5}},  # object of objects or "n/a"
                 SubjectArtefactDescription: str | None = None,
                 ):
        self.InstitutionAddress = InstitutionAddress
        self.InstitutionalDepartmentName = InstitutionalDepartmentName
        self.Manufacturer = Manufacturer
        self.ManufacturersModelName = ManufacturersModelName
        self.SoftwareVersions = SoftwareVersions
        self.TaskDescription = TaskDescription
        self.Instructions = Instructions
        self.CogAtlasID = CogAtlasID
        self.CogPOID = CogPOID
        self.DeviceSerialNumber = DeviceSerialNumber
        self.EEGReference = EEGReference
        self.SamplingFrequency = SamplingFrequency
        self.PowerLineFrequency = PowerLineFrequency
        self.SoftareFilters = SoftwareFilters
        self.CapManufacturer = CapManufacturer
        self.CapManufacturersModelName = CapManufacturersModelName
        self.EEGChannelCount = EEGChannelCount
        self.ECGChannelCount = ECGChannelCount
        self.EMGChannelCount = EMGChannelCount
        self.EOGChannelCount = EOGChannelCount
        self.MiscChannelCount = MiscChannelCount
        self.TriggerChannelCount = TriggerChannelCount
        self.RecordingDuration = RecordingDuration
        self.RecordingType = RecordingType
        self.EpochLength = EpochLength
        self.EEGGround = EEGGround
        self.HeadCircumference = HeadCircumference
        self.EEGPlacementScheme = EEGPlacementScheme
        self.HardwareFilters = HardwareFilters
        self.SubjectArtefactDescription = SubjectArtefactDescription

    def to_json_file(self,
                     file_path: str | None = None):
        json_buffer = json.dumps(self.__dict__, indent=2)
        with open(file_path,
                  mode="w+",
                  encoding="utf-8") as _f:
            _f.write(json_buffer)
        return json_buffer


class BDFParameters(object):
    def __init__(self,
                 data: u.Quantity,
                 channel_labels: List[str] = ['LeftMastoid', 'RightMastoid'],
                 transducer_label: str = 'passive electrode',
                 pre_filtering_label: str = 'HP:3; LP:410',
                 fs: type(u.quantity) | None = None,
                 events: Events | None = None,
                 triggers_code_table: pd.DataFrame | None = None,
                 recording_id: str = '',  # 80 bytes
                 start_date: str = 'dd.mm.yy',
                 start_time: str = 'hh.mm.ss',
                 physical_minimum_quantity: type(u.quantity) | None = None,
                 physical_maximum_quantity: type(u.quantity) | None = None):
        self.data = data
        self.channel_labels = channel_labels
        self.transducer_label = transducer_label
        self.pre_filtering_label = pre_filtering_label
        self.fs = fs
        self.events = events
        self.triggers_code_table = triggers_code_table
        self.recording_id = recording_id
        self.start_date = start_date
        self.start_time = start_time
        self.physical_minimum_quantity = physical_minimum_quantity
        self.physical_maximum_quantity = physical_maximum_quantity


def events_to_bids(events: Events,
                   triggers_code_table: pd.DataFrame,
                   exclude_code: List[float] | None = None):
    """
        Convert events to BIDS compatible events pandas dataframe
        :param events: Events class containing all the events
        :param triggers_code_table: pandas dataframe. It must contain the trigger_code column and any additional columns
        to be included in the output data_frame
        :param exclude_code: list of codes to be ignored
        :return: pandas dataframe ensuring onset and duration columns as required by BIDS standard.
    """
    events_pd = events.events_to_pandas(exclude_code=exclude_code)
    events_pd = events_pd.rename(columns={"time_pos": "onset",
                                          "dur": "duration",
                                          "code": "trigger_code"
                                          })
    events_pd = events_pd.merge(triggers_code_table, on='trigger_code')
    return events_pd


class BIDSEventAnnotation(object):
    def __init__(self,
                 variable_name: str | None = None,
                 LongName: str | None = None,
                 Description: str | None = None,
                 Units: str | None = None,
                 Levels: dict | None = None,
                 HED: {str} = None):
        self.variable_name = variable_name
        self.LongName = LongName
        self.Description = Description
        self.Units = Units
        self.Levels = Levels
        self.HED = HED

    def get_name_and_values(self):
        name = self.variable_name
        values = {'LongName': self.LongName,
                  'Description': self.Description,
                  'Units': self.Units,
                  'Levels': self.Levels,
                  'HED': self.HED,
                  }
        return name, values


class GeneratedBy(object):
    def __init__(self,
                 Name: str | None = None,  # Name of the pipeline or process that generated the outputs. Use "Manual" to
                 # indicate the derivatives were generated by hand, or adjusted manually after an initial run of an
                 # automated pipeline..
                 Version: str | None = None,  # Version of the pipeline
                 Description: str | None = None,  # Plain-text description of the pipeline or process that generated the
                 # outputs. RECOMMENDED if Name is "Manual".
                 CodeURL: str | None = None,  # URL where the code used to generate the dataset may be found.
                 Container: List[dict] | None = None,  # Used to specify the location and relevant attributes of
                 # software
                 # container image used to produce the dataset. Valid keys in this object include Type, Tag and URI
                 # with string values.
                 ):
        self.Name = Name
        self.Version = Version,
        self.Description = Description
        self.CodeURL = CodeURL
        self.Container = Container


class BIDSDataSetDescription(object):
    def __init__(self,
                 Name: str = "",  # Name of the dataset.
                 BIDSVersion: str = '1.10.0',  # The version of the BIDS standard that was used.
                 HEDVersion: str | None = None,  # If HED tags are used: The version of the HED schema used to validate
                 # HED tags for study. May include a single schema or a base schema and one or more library schema.
                 DatasetLinks: {str} = None,  # DatasetLinks Used to map a given <dataset-name> from a BIDS URI of the
                 # form bids:<dataset-name>:path/within/dataset to a local or remote location. The <dataset-name>: ""
                 # (an empty string) is a reserved keyword that MUST NOT be a key in DatasetLinks (example:
                 # bids::path/within/dataset).
                 DatasetType: str = 'raw',  # The interpretation of the dataset. For backwards compatibility,
                 # the default value is "raw".
                 License: str = "",  # The license for the dataset. The use of license name abbreviations is
                 # RECOMMENDED for specifying a license (see Licenses). The corresponding full license text MAY be
                 # specified in an additional LICENSE file.
                 Authors: List[str] = [""],   # 'List of individuals who contributed to the creation/curation of
                 # the dataset.'
                 HowToAcknowledge: str = "",  # Text containing instructions on how researchers using this
                 # dataset should acknowledge the original authors. This field can also be used to define a publication
                 # that should be cited in publications that use the dataset.
                 Funding: List[str] = [""],  # List of sources of funding (grant numbers).
                 EthicsApprovals: List[str] = [""],  # List of ethics committee approvals of the research protocols
                 # and/or protocol identifiers.
                 ReferencesAndLinks: List[str] = [""],  # List of references to publications that contain information
                 # on the dataset. A reference may be textual or a URI.
                 DatasetDOI: str = "",  # The Digital Object Identifier of the dataset (not the corresponding paper).
                 # DOIs SHOULD be expressed as a valid URI; bare DOIs such as 10.0.2.3/dfjj.10 are DEPRECATED.
                 GeneratedBy: List[GeneratedBy] = [{}],  # Used to specify provenance of the dataset.
                 SourceDatasets: List[str] = [""],  # Used to specify the locations and relevant attributes of all
                 # source datasets. Valid keys in each object include "URL", "DOI" (see URI), and "Version" with
                 # string values.
                 ):
        self.Name = Name
        self.BIDSVersion = BIDSVersion
        self.HEDVersion = HEDVersion
        self.DatasetLinks = DatasetLinks
        self.DatasetType = DatasetType
        self.License = License
        self.Authors = Authors
        self.HowToAcknowledge = HowToAcknowledge
        self.Funding = Funding
        self.EthicsApprovals = EthicsApprovals
        self.ReferencesAndLinks = ReferencesAndLinks
        self.DatasetDOI = DatasetDOI
        self.GeneratedBy = GeneratedBy
        self.SourceDatasets = SourceDatasets
    def to_json_file(self,
                     file_path: str | None = None):
        json_buffer = json.dumps(self.__dict__, indent=2)
        with open(file_path,
                  mode="w+",
                  encoding="utf-8") as _f:
            _f.write(json_buffer)
        return json_buffer


class BIDSParticipant(object):
    def __init__(self,
                 participant_id: str | None = None,  # A participant identifier of the form sub-<label>, matching a
                 # participant entity found in the dataset. There MUST be exactly one row for each participant.
                 # Values in participant_id MUST be unique. This column must appear first in the file.
                 species: (str | float) = 'homo sapiens',  # The species column SHOULD be a binomial species name from
                 # the NCBI Taxonomy (for example, homo sapiens, mus musculus, rattus norvegicus).
                 # For backwards compatibility, if species is absent, the participant is assumed to be homo sapiens.
                 age: (float | int) = None,  # Numeric value in years (float or integer value). It is RECOMMENDED to tag
                 # participant ages that are 89 or higher as 89+, for privacy purposes.
                 sex: str | None = None,  # String value indicating phenotypical sex, one of "male", "female", "other".
                 # For "male", use one of these values: male, m, M, MALE, Male.
                 # For "female", use one of these values: female, f, F, FEMALE, Female.
                 # For "other", use one of these values: other, o, O, OTHER, Other.
                 # This column may appear anywhere in the file.
                 handedness: str | None = None,  # String value indicating one of "left", "right", "ambidextrous".
                 # For "left", use one of these values: left, l, L, LEFT, Left.
                 # For "right", use one of these values: right, r, R, RIGHT, Right.
                 # For "ambidextrous", use one of these values: ambidextrous, a, A, AMBIDEXTROUS, Ambidextrous.
                 # This column may appear anywhere in the file.
                 strain: (str | float) = None,  # For species different from homo sapiens, string value indicating the
                 # strain of the species, for example: C57BL/6J.
                 strain_rrid: str | None = None,  # For species different from homo sapiens, research resource
                 # identifier (RRID) of the strain of the species, for example: RRID:IMSR_JAX:000664.
                 # This column may appear anywhere in the file.
                 additional_columns: dict = None  # dictionary with additional columns to include
                 ):
        self.participant_id = participant_id
        self.species = species
        self.age = age
        self.sex = sex
        self.handedness = handedness
        self.strain = strain
        self.strain_rrid = strain_rrid
        if additional_columns is not None:
            for attr, value in additional_columns.items():
                setattr(self, attr, value)

    def save_as_tvs(self, output_path: str | None = None):
        all_participants_df = pd.DataFrame()
        if os.path.exists(output_path):
            all_participants_df = pd.read_csv(filepath_or_buffer=output_path,
                                              sep='\t',
                                              dtype=str)
        current_participant_df = pd.DataFrame([self.__dict__])
        if ((all_participants_df.shape[0] > 0) and
                all_participants_df['participant_id'].isin([self.participant_id]).any()):
            return

        all_participants_df = pd.concat([all_participants_df, current_participant_df])
        all_participants_df.to_csv(
            path_or_buf=output_path,
            sep='\t',
            index=False)


class BIDSParticipantAnnotation(object):
    def __init__(self,
                 variable_name: str | None = None,
                 Description: str | None = None,
                 Units: str | None = None,
                 Levels: dict | None = None,
                 HED: {str} = None):
        self.variable_name = variable_name
        self.Description = Description
        self.Units = Units
        self.Levels = Levels

    def get_name_and_values(self):
        name = self.variable_name
        values = {'Description': self.Description,
                  'Units': self.Units,
                  'Levels': self.Levels
                  }
        return name, values


class BIDSEventDescriptionItem(list):
    def append(self, item):
        # optional processing here
        assert isinstance(item, BIDSEventAnnotation)
        super(BIDSEventDescriptionItem, self).append(item)

    def save_to_json(self, output_path: str | None = None):
        _dict = {}
        for _e in self:
            _name, _value = _e.get_name_and_values()
            _dict[_name] = _value
        json_buffer = json.dumps(_dict, indent=2)
        with open(output_path,
                  mode="w+",
                  encoding="utf-8") as _f:
            _f.write(json_buffer)
        return json_buffer


class BIDSParticipantDescriptionItem(list):
    def append(self, item):
        # optional processing here
        assert isinstance(item, BIDSParticipantAnnotation)
        super(BIDSParticipantDescriptionItem, self).append(item)

    def save_to_json(self, output_path: str | None = None):
        _dict = {}
        for _e in self:
            _name, _value = _e.get_name_and_values()
            _dict[_name] = _value
        json_buffer = json.dumps(_dict, indent=2)
        with open(output_path,
                  mode="w+",
                  encoding="utf-8") as _f:
            _f.write(json_buffer)
        return json_buffer


def check_allowed_patterh(str, pattern):
    # _matching the strings
    valid = False
    if re.search(pattern, str):
        valid = True
    return valid


class BIDSConverter(object):
    """
    This class implement several functionality to facilitate BIDS format exportation.
    https://bids-specification.readthedocs.io/en/stable/modality-specific-files/electroencephalography.html
    """
    def __init__(self,
                 output_path: Path | None = None,
                 current_eeg_file: Path | None = None,
                 bdf_parameters: BDFParameters | None = None,
                 dataset_description: BIDSDataSetDescription | None = None,
                 participant: BIDSParticipant = BIDSParticipant(),
                 eeg_json: BIDS_EEG_JSON | None = None,
                 event_column_descriptions: BIDSEventDescriptionItem = BIDSEventDescriptionItem(),
                 participant_column_descriptions: BIDSParticipantDescriptionItem = BIDSParticipantDescriptionItem(),
                 readme: str = """This is a great study with cool experiments and
                                                   excellent data"""):
        """
        Initializes the path where BIDS format will be saved
        :param output_path: path to save data using BIDS format
        :param current_eeg_file: path with current EEG data file
        :param bdf_parameters: BDFParameters class used to generate BDF files. If BDF files exists, this should not
        be used.
        :param dataset_description: BIDSDataSetDescription class with description following the BIDS standard
        :param session_id: the session ID of the current EEG file as defined in the BIDS standard
        :param task_id: : the task ID of the current EEG file
        :param participant: BIDSParticipant class with participant information as defined in the BIDS standard
        :param eeg_json: BIDS_EEG_JSON class with EEG_json information of current EEG file as defined in BIDS standard
        :param event_column_descriptions: BIDSEventDescriptionItem class defnining events as required by the BIDS
        standard
        :param readme: string with the description of the data set.
        """
        self.output_path = output_path
        self.current_eeg_file = current_eeg_file
        self.bdf_parameters = bdf_parameters
        self.dataset_description = dataset_description
        self.eeg_json = eeg_json
        self.event_column_descriptions = event_column_descriptions
        self.participant_column_descriptions = participant_column_descriptions
        self.readme = readme
        self._subject_id = None
        self._session_id = None
        self._task_id = None
        self._run_id = None

    def get_subject_id(self):
        return self._subject_id

    def set_subject_id(self, value):
        pattern = re.compile('^sub-[0-9a-zA-Z]+$')
        assert check_allowed_patterh(value, pattern), 'the label is not in compliance with BIDS (sub-[0-9a-zA-Z])'
        self._subject_id = value
    subject_id = property(get_subject_id, set_subject_id)

    def get_session_id(self):
        return self._session_id

    def set_session_id(self, value):
        pattern = re.compile('^ses-[0-9a-zA-Z]+$')
        assert check_allowed_patterh(value, pattern), 'the label is not in compliance with BIDS (ses-[0-9a-zA-Z])'
        self._session_id = value
    session_id = property(get_session_id, set_session_id)

    def get_task_id(self):
        return self._task_id

    def set_task_id(self, value):
        pattern = re.compile('^task-[0-9a-zA-Z]+$')
        assert check_allowed_patterh(value, pattern), 'the label is not in compliance with BIDS (task-[0-9a-zA-Z])'
        self._task_id = value
    task_id = property(get_task_id, set_task_id)

    def get_run_id(self):
        return self._run_id

    def set_run_id(self, value):
        pattern = re.compile('^run-[0-9]+$')
        assert check_allowed_patterh(value, pattern), 'the label is not in compliance with BIDS (run-[0-9])'
        self._run_id = value
    run_id = property(get_run_id, set_run_id)

    def _get_current_path(self):
        out = self.output_path.joinpath(
            self.subject_id).joinpath(
            self.session_id
        ).joinpath('eeg')
        return out

    def _get_current_basename(self):
        out = (self.subject_id +
               '_' +
               self.session_id +
               '_' +
               self.task_id +
               '_' +
               self.run_id +
               '_'
               )
        return out

    def save_eeg_data_as_bdf(self):
        output_file_name = self._get_current_path().joinpath(self._get_current_basename() + 'eeg.bdf')
        data_to_bdf(output_file_name=output_file_name,
                    data=self.bdf_parameters.data,
                    channel_labels=self.bdf_parameters.channel_labels,
                    transducer_label=self.bdf_parameters.transducer_label,
                    pre_filtering_label=self.bdf_parameters.pre_filtering_label,
                    fs=self.bdf_parameters.fs,
                    events=self.bdf_parameters.events,
                    subject_id=self.subject_id,
                    recording_id=self.bdf_parameters.recording_id,
                    start_date=self.bdf_parameters.start_date,
                    start_time=self.bdf_parameters.start_time,
                    physical_minimum_quantity=self.bdf_parameters.physical_minimum_quantity,
                    physical_maximum_quantity=self.bdf_parameters.physical_maximum_quantity
                    )
        self.current_eeg_file = output_file_name

    def save_eeg_events_data(self,
                             triggers_code_table: pd.DataFrame | None = None,
                             exclude_event_code: float = 0):
        # extract EEG data ####
        output_file_name = self._get_current_path().joinpath(self._get_current_basename() + 'eeg.bdf')
        if not os.path.exists(str(output_file_name)):
            shutil.copy(self.current_eeg_file, output_file_name)

        reader = ReadInputData(file_path=output_file_name)
        reader.run()

        events_tsv = events_to_bids(events=reader.output_node.events,
                                    exclude_code=exclude_event_code,
                                    triggers_code_table=self.bdf_parameters.triggers_code_table)
        events_tsv.to_csv(
            path_or_buf=self._get_current_path().joinpath(self._get_current_basename() + 'events.tsv'),
            sep='\t',
            index=False)

        self.event_column_descriptions.save_to_json(
            output_path=self._get_current_path().joinpath(self._get_current_basename() + 'events.json'))

        self.bdf_parameters.triggers_code_table.to_csv(
            path_or_buf=self._get_current_path().joinpath(self._get_current_basename() + 'triggerCodes.tsv'),
            sep='\t',
            index=False)

        self.eeg_json.SamplingFrequency = reader.output_node.fs.to('Hz').value,
        self.eeg_json.EEGChannelCount = reader.output_node.data.shape[1],
        self.eeg_json.RecordingDuration = reader.output_node.data.shape[0] / reader.output_node.fs.to('Hz').value,
        self.eeg_json.to_json_file(
            file_path=self._get_current_path().joinpath(self._get_current_basename() + 'eeg.json'))

    def readme_to_md(self, file_path: str | None = None):
        with open(file_path, "w+") as _file:
            _file.write("Dataset description \n")
            _file.write("===================== \n")
            _file.write(self.readme)

    def save_current_iteration(self):
        os.makedirs(self._get_current_path(), exist_ok=True)
        if self.bdf_parameters is not None:
            self.save_eeg_data_as_bdf()
        self.save_eeg_events_data(exclude_event_code=0)
        self.participant.save_as_tvs(self.output_path.joinpath('participants.tsv'))
        self.participant_column_descriptions.save_to_json(output_path=self.output_path.joinpath('participants.json'))
        self.dataset_description.to_json_file(self.output_path.joinpath('dataset_description.json'))
        self.readme_to_md(self.output_path.joinpath('README.md'))



import numpy as np
import peegy.processing.tools.weightedAverage as averager
import peegy.io.external_tools.aep_gui.medel_stimuli_reader as medel_reader
from peegy.definitions.eeg_definitions import signal_information
__author__ = 'jundurraga-ucl'


def generate_frequency_dict(j_averager=averager.JAverager, channels=[{}]):
    out = []
    _j_h_test = j_averager.hotelling_t_square_test()
    for i, ch in enumerate(channels):
        for j, h_test in enumerate(_j_h_test[i]):
            out.append(dict(ch, **h_test))
    return out


def get_frequencies_of_interest(stimulus_parameters=[signal_information()]):
    # trigger_channel = parameters['Measurement']['MeasurementModule']['TriggerChannel']
    freq = np.array([], dtype=np.float64)
    for i, _parameters in enumerate(stimulus_parameters):
        if _parameters['type'] in ['generateSinusoidalNotchNoise', 'sinusoidal']:
            freq = np.append(freq, _parameters['ModulationFrequency'])
            freq = np.append(freq, _parameters['CarrierFrequency'])
        elif _parameters['type'] in ['sin_alt_phase', 'medelAM_Alt_Phase_BiphasicPulses', 'transposeToneAltPhase']:
            freq = np.append(freq, [_parameters['AltPhaseRate'] * 2.0 * i for i in np.arange(1, 4)])
        elif _parameters['type'] in ['generateNoiseAltDelaySignal']:
            if 'ITDModRate' in list(_parameters.keys()):
                s_rate = _parameters['ITDModRate']
            if 'NITDModulations' in list(_parameters.keys()):
                s_rate = _parameters['NITDModulations'] / _parameters['Duration']
            freq = np.append(freq, _parameters['ModulationFrequency'])
            freq = np.append(freq, [s_rate * i for i in np.arange(1, 4)])
        elif _parameters['type'] in ['sin_alt_burst']:
            s_rate = _parameters['AltLevelRate']
            freq = np.append(freq, _parameters['ModulationFrequency'])
            freq = np.append(freq, [s_rate * i for i in np.arange(1, 4)])
            freq = np.append(freq, _parameters['ModulationFrequency'] + s_rate)
            freq = np.append(freq, _parameters['ModulationFrequency'] - s_rate)
        elif _parameters['type'] in ['attentional_sin']:
            freq = np.append(freq, _parameters['ModulationFrequencyTa'])
            freq = np.append(freq, _parameters['ModulationFrequencyTb'])
            freq = np.append(freq, _parameters['ModulationFrequencyDa'])
            freq = np.append(freq, _parameters['ModulationFrequencyDb'])
            freq = np.append(freq, _parameters['CarrierFrequencyTa'])
            freq = np.append(freq, _parameters['CarrierFrequencyTb'])
            freq = np.append(freq, _parameters['CarrierFrequencyDa'])
            freq = np.append(freq, _parameters['CarrierFrequencyDb'])
        elif 'AltLevelRate' in _parameters:
            s_rate = _parameters['AltLevelRate']
            freq = np.append(freq, _parameters['ModulationFrequency'])
            freq = np.append(freq, [s_rate * i for i in np.arange(1, 4)])
            freq = np.append(freq, _parameters['ModulationFrequency'] + s_rate)
            freq = np.append(freq, _parameters['ModulationFrequency'] - s_rate)
        elif 'IPMRate' in _parameters:
            s_rate = _parameters['IPMRate']
            freq = np.append(freq, _parameters['ModulationFrequency'])
            freq = np.append(freq, [s_rate * i for i in np.arange(1, 4)])
            freq = np.append(freq, _parameters['ModulationFrequency'] + s_rate)
            freq = np.append(freq, _parameters['ModulationFrequency'] - s_rate)
        elif 'ITMRate' in _parameters:
            s_rate = _parameters['ITMRate']
            freq = np.append(freq, _parameters['ModulationFrequency'])
            freq = np.append(freq, [s_rate * i for i in np.arange(1, 4)])
            freq = np.append(freq, _parameters['ModulationFrequency'] + s_rate)
            freq = np.append(freq, _parameters['ModulationFrequency'] - s_rate)
            if 'Rate' in _parameters:
                freq = np.append(freq, _parameters['Rate'])
        elif 'AltRate' in _parameters and 'ModulationFrequency_1' in _parameters and\
                'ModulationFrequency_2' in _parameters:
            s_rate = _parameters['AltRate']
            freq = np.append(freq, [s_rate * i for i in np.arange(1, 4)])
            freq = np.append(freq, s_rate / 2.0)
            freq = np.append(freq, _parameters['ModulationFrequency_1'])
            freq = np.append(freq, _parameters['ModulationFrequency_2'])
        freq = np.delete(freq, np.where(freq == 0)[0])
    return np.unique(freq)


def get_stimulus_parameters(parameters={}):
    stimuli = []
    if isinstance(parameters, dict) and 'Measurement' in list(parameters.keys()):
        trigger_stimulus = parameters['Measurement']['MeasurementModule']['TriggerChannel']
        for i, stimulus in enumerate(parameters['Measurement']['StimuliModule']['Stimulus']):
            if not stimulus['Parameters']['SignalType'] in ['medelAM_Alt_Phase_BiphasicPulses',
                                                            'medelAMBiphasicPulsesElectChange',
                                                            'generateNoiseAltDelaySignal']:
                if stimulus['Parameters']['Channel'] == trigger_stimulus['Parameters']['Channel']:
                    continue
            _signal = signal_information(**{'type': stimulus['Parameters']['SignalType']})
            for _key in list(stimulus['Parameters'].keys()):
                _signal[_key] = stimulus['Parameters'][_key]
            stimuli.append(_signal)
    return stimuli


def get_interpolation_data_points(**kwargs):
    parameters = kwargs.get('parameters', {})
    width = kwargs.get('width', 1200.0e-6)
    fs = kwargs.get('width', 16384.0)
    _interpolation_points = np.array([])
    for i, stimulus in enumerate(parameters['Measurement']['StimuliModule']['Stimulus']):
        stimulus_parameters = stimulus['Parameters']
        if stimulus_parameters['SignalType'] in ['medelAM_Alt_Phase_BiphasicPulses']:
            _interpolation_points = np.concatenate((_interpolation_points,
                                                    medel_reader.get_medel_am_alt_phase_biphasic_pulse_positions(
                                                        stimulus_parameters, **kwargs)))
            _interpolation_points = np.unique(_interpolation_points)
        if stimulus_parameters['SignalType'] in ['medelAM_BiphasicPulses', 'medelAMBiphasicPulsesElectChange']:
            _interpolation_points = np.concatenate((_interpolation_points,
                                                    medel_reader.get_medel_am_biphasic_pulse_positions(
                                                        stimulus_parameters, **kwargs)))
        _interpolation_points = _interpolation_points.astype(int)
    interpolation_points = {'ini': _interpolation_points,
                            'end': _interpolation_points + np.round(width * fs).astype(int)}
    return interpolation_points


def cat_dictionary_list(dict_list=[{}]):
    out = {}
    for i, _dict in enumerate(dict_list):
        c_out = {}
        keys = _dict.keys()
        for j, _key in enumerate(keys):
            c_out[_key + '_{}'.format(i)] = _dict[_key]
        out = dict(out, **c_out)
    return out



import os
from os.path import sep
import shutil
from peegy.definitions.default_paths import test_data_path


class DirectoryPaths(object):
    def __init__(self,
                 file_path: str = '',
                 delete_all: bool = False,
                 delete_figures: bool = False,
                 figures_folder: str | None = None,
                 figures_subset_folder=''):
        file_path = os.path.expanduser(file_path)
        self._file_path = file_path
        self._file_directory = os.path.dirname(os.path.abspath(file_path))
        self.file_basename_path = os.path.abspath(os.path.splitext(file_path)[0])
        self.delete_all = delete_all
        self.delete_figures = delete_figures
        self._figures_folder = figures_folder
        if self._figures_folder is None:
            self._figures_folder = os.path.join(self._file_directory, 'figures')
        if (self.delete_all or self.delete_figures) and os.path.exists(self._figures_folder):
            try:
                shutil.rmtree(self._figures_folder)
            except OSError:
                print((OSError.message))
        self.data_dir = None
        self.figures_subset_folder = figures_subset_folder
        self._test_data_path = None

    def set_file_path(self, value):
        self._file_path = value

    def get_file_path(self):
        if not os.path.exists(self._file_path):
            os.makedirs(self._file_path)
        return self._file_path

    file_path = property(get_file_path, set_file_path)

    def set_file_directory(self, value):
        self._file_directory = value

    def get_file_directory(self):
        if not os.path.exists(self._file_directory):
            os.makedirs(self._file_directory)
        return self._file_directory

    file_directory = property(get_file_directory, set_file_directory)

    def get_figures_folder(self):
        if not os.path.exists(self._figures_folder):
            os.makedirs(self._figures_folder)
        return self._figures_folder

    def set_figures_folder(self, value):
        self._figures_folder = value

    figures_folder = property(get_figures_folder, set_figures_folder)

    def get_figure_basename_path(self):
        return os.path.join(self.figures_current_dir, os.path.basename(self.file_path).split('.')[0])

    figure_basename_path = property(get_figure_basename_path)

    def get_data_basename_path(self):
        return os.path.join(self.data_subset_path, os.path.basename(self.file_path).split('.')[0])

    data_basename_path = property(get_data_basename_path)

    def get_figures_current_dir(self):
        _path = os.path.join(self._figures_folder, self.figures_subset_folder) + sep
        if not os.path.exists(_path):
            os.makedirs(_path)
        return _path

    figures_current_dir = property(get_figures_current_dir)

    def get_data_subset_path(self):
        _path = os.path.join(self.data_dir, self.figures_subset_folder)
        if not os.path.exists(_path):
            os.makedirs(_path)
        return _path

    data_path = property(get_data_subset_path)

    def set_test_path(self, value: str | None = None):
        self._test_data_path = value

    def get_test_path(self):
        if self._test_data_path is None:
            self._test_data_path = test_data_path
        if not os.path.exists(self._test_data_path):
            os.makedirs(self._test_data_path)
        return self._test_data_path

    test_path = property(get_test_path, set_test_path)



# -*- coding: utf-8 -*-
import os
from peegy.io.external_tools.aep_gui.dataReadingTools import get_files_and_meta_data
from peegy.io.external_tools.aep_gui.extsys_tools import extsys_to_json
import matplotlib
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')

if __name__ == "__main__":
    eeg_data_directory = r'/home/jundurraga/Documents/Measurements/peegy_data/IPM-FR/'
    to_process = get_files_and_meta_data(eeg_data_directory)

    for _, _data_links in to_process.iterrows():
        _, _name = os.path.split(_data_links.data_links.data_file)
        extsys_to_json(input_file_name=_data_links.data_links.parameters_file,
                       output_file_name=_name,
                       trash_original=True)



# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
from peegy.io.eeg.reader import eeg_reader
from peegy.io.exporters.edf_bdf_writer import split_bdf_by_event_code
from peegy.io.external_tools.file_tools import get_files_and_meta_files
from peegy.io.external_tools.aep_gui.dataReadingTools import get_files_and_meta_data
import numpy as np
from peegy.processing.events.event_tools import get_events
import os
from pathlib import Path
import matplotlib
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')


_path = Path(os.path.abspath(os.path.dirname(__file__)))
data_folder = _path.parent.parent.absolute().joinpath("test_data/")
data_folder = r'/home/jundurraga/Documents/Measurements/ITD-FR-Pilots/JU/'
data_files_meta = get_files_and_meta_files(data_folder, file_types=['bdf'])
_first_file = data_files_meta.iloc[0].data_links.data_file
output_file_path = _path.parent.parent.absolute().joinpath('split_test')
split_bdf_by_event_code(input_file_name=_first_file,
                        event_code=16.0,
                        trash_original=True)

data_files_meta = get_files_and_meta_data(data_folder)
_first_file = data_files_meta.iloc[0].data_links.data_file
data4 = eeg_reader(file_name=_first_file)
data_ds, events_ds, _, _ = data4.get_data()
et_ds = get_events(event_channel=events_ds, fs=data4.fs)
et_ds.summary()
fig1 = plt.figure(figsize=(7, 6), dpi=100)
t_ds = np.arange(data_ds.shape[0]) / data4.fs
ax = fig1.add_subplot(1, 1, 1)
ax.plot(t_ds, data_ds[:, 0])
ax.plot(t_ds, events_ds)
plt.show()



import pandas as pd
import numpy as np
import astropy.units as u
from peegy.definitions.channel_definitions import ChannelItem
import os.path
import datetime
from peegy.tools.units.unit_tools import set_default_unit
from typing import List


def read_header(file_path: str | None = None,
                gain_col_name: str | None = None,
                fs_col_name: str | None = None,
                fs_unit: u.quantity.Quantity = u.Hz,
                gain_unit: u.quantity.Quantity = u.uV,
                gain_inverted: bool = False
                ) -> (dict, np.array):
    """
    Read header from cvs file. This is assumed to be the header indicating the parameters of the data.
    For the moment we assume one data channel, each row being an epoch.
    :param file_path: path to text file
    :param gain_col_name: string indicating the name of the column where data scaling factor should be obtained
    :param fs_col_name: string indicating the name of the column where data sampling rate should be obtained
    :param fs_unit: unit of sampling rate
    :param gain_unit: units of data
    :param gain_inverted: if True, gain will first be inverted and then multiplied
    :return: header information
    """
    header = pd.read_csv(file_path, sep="\t", header=0, nrows=1)
    # add extra information to make compatible with peegy-python
    header = header.to_dict('records')[0]
    header['n_channels'] = 1
    header['fs'] = (1.0 if fs_col_name is None else header[fs_col_name]) * fs_unit
    # header['gain'] = header['Volt/bit(nV)']
    header['gain'] = (1.0 if gain_col_name is None else header[gain_col_name]) * gain_unit
    if gain_inverted:
        header['gain'] = 1 / header['gain']
        gain_unit = 1 / gain_unit
    header['channels'] = np.array([ChannelItem(label='CH_0', idx=0)])
    header['file_name'] = file_path
    _start_date_format = "%d.%m.%y"
    _start_time_format = "%H.%M.%S"
    struct_data_time = datetime.datetime.strptime(
        datetime.datetime.fromtimestamp(os.path.getctime(file_path)).strftime(_start_date_format + _start_time_format),
        _start_date_format + _start_time_format)
    header['start_date'] = struct_data_time.strftime(_start_date_format)
    header['start_time'] = struct_data_time.strftime(_start_time_format)
    return header


def read_channel(header: dict | None = None,
                 channels_idx: np.array = np.array([]),
                 ini_time: float = 0,
                 end_time: float | None = None,
                 from_row: int = 2,
                 ) -> np.array:
    """
    Read data from csv files. For now we assumed single channel data, one row per epoch.
    :param header: dictionary containing the header of the file
    :param channels_idx: indexes of channels to be read
    :param ini_time: time (in seconds) to read from
    :param end_time: time (in seconds) to read up to
    :param from_row: integer indicating in which row data start.
    :return: raw data
    """
    ini_time = set_default_unit(ini_time, u.s)
    end_time = set_default_unit(end_time, u.s)
    data = pd.read_csv(header['file_name'], sep="\t", skiprows=from_row,  header=None)
    data = data.iloc[:, 1::].to_numpy()
    gain = header['gain']
    data = data * gain
    events = np.zeros(data.shape)
    events[:, 0] = 1
    data = data.reshape(-1, 1)
    events = events.reshape(-1, 1)
    _ini_pos = np.minimum(np.maximum(0, ini_time * header['fs']), data.shape[0]).astype(int)
    if end_time is None:
        _end_pos = data.shape[0]
    else:
        _end_pos = np.minimum(np.maximum(0, end_time * header['fs']), data.shape[0]).astype(int)
    if not channels_idx.size:
        channels_idx = np.arange(data.shape[1])
    data = data[_ini_pos: _end_pos, channels_idx]
    events = events[_ini_pos: _end_pos, :]
    return data, events, data.unit


def read_file(file_path: str | None = None) -> (pd.DataFrame, np.array):
    """
    Read data from cvs file
    :param file_path: path to text file
    :return: header information and raw data
    """
    header = pd.read_csv(file_path, sep="\t", header=0, nrows=1)
    data = pd.read_csv(file_path, sep="\t", skiprows=2,  header=None)
    data = data.iloc[:, 1::].to_numpy()
    data = data.T[:, None, :]
    return header, data


def merge_files(files_path: List[str]):
    data = None
    headers = pd.DataFrame()
    for _file in files_path:
        _header, _data = read_file(file_path=_file)
        headers = pd.concat([headers, _header], ignore_index=True)
        if data is None:
            data = _data
        else:
            data = np.dstack((data, _data))
    if not np.all(headers.all()):
        RuntimeError('Files must have same header to be merged')
    header = headers.iloc[0, :]
    return header, data



from peegy.processing.pipe.definitions import InputOutputProcess
import matplotlib.pyplot as plt
import numpy as np
import os
from matplotlib.ticker import FormatStrFormatter
from PyQt5.QtCore import QLibraryInfo
from peegy.tools.units.unit_tools import set_default_unit
import astropy.units as u
from matplotlib.ticker import AutoMinorLocator

os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = QLibraryInfo.location(QLibraryInfo.PluginsPath)


class InterpolateData(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 ini_end_points: np.array = np.array([[], []]),
                 save_plots: bool = True,
                 n_plots: int = 9,
                 fig_format: str = '.png',
                 fontsize: float = 8,
                 idx_channels_to_plot: np.array = np.array([0]),
                 user_naming_rule: str = 'interpolation',
                 **kwargs):
        """
        This class will interpolate data on each channel at a between two points passed in ini_end_points
        :param input_process: an InputOutput process
        :param ini_end_points: a Nx2 numpy array with the interpolation points
        :param save_plots: bool, if true, figures will be generated and saved showing a zoomed progression of the interp
        :param n_plots: int: number of plots (with different zoom scales) to be generated
        :param fig_format: string indicating the format of the output figure (e.g. '.png' or '.pdf')
        :param fontsize: size of fonts in plot
        :param idx_channels_to_plot: np.array indicating the index of the channels to plot
        :param user_naming_rule: string indicating a user naming to be included in the figure file name
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(InterpolateData, self).__init__(input_process=input_process, **kwargs)
        self.ini_end_points = ini_end_points
        self.save_plots = save_plots
        self.n_plots = n_plots
        self.fig_format = fig_format
        self.fontsize = fontsize
        self.idx_channels_to_plot = idx_channels_to_plot
        self.user_naming_rule = user_naming_rule
        self.fontsize = fontsize

    def transform_data(self):
        if self.ini_end_points.size:
            print('Interpolating data')
            idx_delete = np.concatenate((np.where(self.ini_end_points[:, 0] >
                                                  self.input_node.data.shape[0] - 1)[0],
                                         np.where(self.ini_end_points[:, 1] >
                                                  self.input_node.data.shape[0] - 1)[0]))

            if np.any(idx_delete):
                self.ini_end_points = np.delete(self.ini_end_points, idx_delete, 0)
            self.output_node.data = self.input_node.data.copy()
            for _i in range(self.ini_end_points.shape[0]):
                _ini = self.ini_end_points[_i, 0]
                _end = self.ini_end_points[_i, 1]
                new_x = np.linspace(_ini, _end, num=_end - _ini + 1).astype(int)
                new_x_ar = np.tile(np.atleast_2d(new_x).T, (1, self.input_node.data.shape[1]))
                self.output_node.data[new_x, :] = ((new_x_ar - _ini) * (self.output_node.data[_end, :] -
                                                                        self.output_node.data[_ini, :]) /
                                                   (_end - _ini) + self.output_node.data[_ini, :])
            if self.save_plots:
                for _idx_channel in self.idx_channels_to_plot:
                    fig, ax = plt.subplots(1, 1)
                    ax.plot(self.input_node.x, self.input_node.data[:, _idx_channel])
                    ax.plot(self.input_node.x, self.output_node.data[:, _idx_channel])
                    ax.plot(self.input_node.x[self.ini_end_points[:, 0]],
                            self.output_node.data[self.ini_end_points[:, 0], _idx_channel],
                            'o', color='k', markersize=2)
                    ax.plot(self.input_node.x[self.ini_end_points[:, 1]],
                            self.output_node.data[self.ini_end_points[:, 1], _idx_channel],
                            'o', color='r', markersize=2)
                    m_point = self.ini_end_points.shape[0] // 2
                    figure_dir_path = self.input_node.paths.figures_current_dir
                    _sep = '_' if self.user_naming_rule is not None else ''
                    _naming_rule = '' if self.user_naming_rule is None else self.user_naming_rule
                    figure_basename = self.name + _sep + _naming_rule
                    for _i in range(self.n_plots):
                        _label = self.input_node.layout[_idx_channel].label
                        _fig_path = os.path.join(figure_dir_path, _label + '_' + figure_basename + '_{:}{:}'.format(
                            _i, self.fig_format))
                        _ini = max(0, m_point - 2 * 2 ** _i)
                        _end = min(self.ini_end_points.shape[0], m_point + 2 * 2 ** _i)
                        ax.set_xlim(self.input_node.x[self.ini_end_points[_ini, 0]].to(self.input_node.x.unit).value,
                                    self.input_node.x[self.ini_end_points[_end, 0]].to(self.input_node.x.unit).value)
                        ax.autoscale_view()
                        ax.set_xlabel('Time [{:}]'.format(self.input_node.x.unit), fontsize=self.fontsize)
                        ax.set_ylabel('Amplitude [{:}]'.format(self.input_node.data.unit), fontsize=self.fontsize)
                        ax.xaxis.set_major_formatter(FormatStrFormatter('%.3f'))
                        ax.xaxis.set_minor_locator(AutoMinorLocator())
                        fig.tight_layout()
                        fig.savefig(_fig_path)
                        print('saving interpolated figures in {:}'.format(_fig_path))
                    plt.close(fig)


class PeriodicInterpolation(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 interpolation_rate: type(u.Quantity) | None = None,
                 interpolation_width: u.Quantity = 0.0 * u.s,
                 interpolation_offset: u.Quantity = 0.0 * u.s,
                 event_code: float = 0.0,
                 user_naming_rule: str = '',
                 save_plots: bool = False,
                 **kwargs):
        """
        This class will interpolate data on each channel at a given rate using as a reference time the position of
        trigger events
        :param input_process: InputOutputProcess Class
        :param interpolation_rate: rate (in Hz) to interpolate data points
        :param interpolation_width: the width (in sec) of the interpolation region
        :param interpolation_offset: an offset (in sec) in reference to trigger events
        :param event_code: the event code that will be use as reference to start interpolating
        :param user_naming_rule: string indicating a user naming to be included in the figure file name
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(PeriodicInterpolation, self).__init__(input_process=input_process, **kwargs)
        self.interpolation_rate = set_default_unit(interpolation_rate, u.Hz)
        self.interpolation_width = set_default_unit(interpolation_width, u.s)
        self.interpolation_offset = set_default_unit(interpolation_offset, u.s)
        self.event_code = event_code
        self.user_naming_rule = user_naming_rule
        self.save_plots = save_plots

    def transform_data(self):
        _events = self.input_node.events.get_events(self.event_code)
        _event_times = [_e.time_pos for _e in _events]

        _ini_points = np.array([])
        # read all events and generate interpolation points in the same unit (seconds)
        for _start, _end in zip(_event_times[:-1], _event_times[1::]):
            if self.interpolation_rate is not None:
                _ini_points = np.concatenate((_ini_points, np.arange(_start.to(u.s).value,
                                                                     _end.to(u.s).value,
                                                                     1. / self.interpolation_rate.to(u.Hz).value)))
            else:
                _ini_points = np.append(_ini_points, _start.to(u.s).value)

        if self.interpolation_rate is not None:
            _ini_points = np.concatenate((_ini_points,
                                          np.arange(_event_times[-1].to(u.s).value,
                                                    self.input_node.data.shape[0] / self.input_node.fs.to(u.Hz).value,
                                                    1. / self.interpolation_rate.to(u.Hz).value)))
        _ini_points = _ini_points * u.s
        _ini_points = _ini_points - self.interpolation_offset
        _end_points = _ini_points + self.interpolation_width
        ini_end_points = (np.array([_ini_points, _end_points]).T * self.input_node.fs.to(u.Hz).value).astype(int)
        interp_data = InterpolateData(input_process=self.input_process,
                                      ini_end_points=ini_end_points,
                                      user_naming_rule=self.user_naming_rule,
                                      save_plots=self.save_plots)
        interp_data.run()
        self.output_node = interp_data.output_node



"""
.. _tut-assr-f-test-sim:

##############################################
ASSR F-Test Example (Simulated)
##############################################

In this example we simulate an ASSR and we investigate the significance of the response by means of an F test that
compares the variance of the frequency of interest with the variance of neighbouring frequency bins

"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#     matplotlib.use('Qt5Agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.definitions import Events, Domain
from peegy.processing.pipe.general import ReferenceData, FilterData, AutoRemoveBadChannels, ReSampling
from peegy.processing.pipe.epochs import AverageEpochs, EpochData
from peegy.processing.pipe.attach import AppendGFPChannel
from peegy.processing.pipe.statistics import FTest
from peegy.processing.pipe.plot import PlotWaveforms, PlotTopographicMap
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.template_generator.auditory_waveforms import aep
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
from peegy.processing.pipe.time_frequency import TimeFrequencyResponse, PlotTimeFrequencyData
import os
import astropy.units as u
import numpy as np
from peegy.io.storage.data_storage_reading_tools import sqlite_tables_to_pandas
from peegy.io.storage.plot_tools import plot_topographic_maps

# %%
# Generate some data
# =================================
# We generate some auditory steady-state response (ASSR)


fs = 400.0 * u.Hz
epoch_length = 4.0 * u.s
epoch_length = np.ceil(epoch_length * fs) / fs  # fit to fs rate
assr_frequency = 6 * u.Hz
assr_frequency = np.ceil(epoch_length * assr_frequency) / epoch_length  # fit to epoch length
test_frequencies = np.array([assr_frequency.value]) * u.Hz
# here we pick some random frequencies to test statistical detection
random_frequencies = np.round(np.unique(np.random.rand(1) * 5), decimals=1) * u.Hz
template_waveform, _ = aep(fs=fs)
n_channels = 32
event_times = np.arange(0, 360.0, 1 / assr_frequency.to(u.Hz).value) * u.s
reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.arange(n_channels)) / n_channels,
                           snr=5,
                           layout_file_name='biosemi32.lay',
                           event_times=event_times,
                           event_code=1.0,
                           figures_subset_folder='assr_f_test',
                           events_asymmetry_factor=0.7,
                           events_asymmetry_interval=2)
reader.run()

# %%
# Resize events
# ============================
# Now we keep events at intervals that correspond to our desired epoch length


events = reader.output_node.events.get_events(code=1)
# skip events to preserve only those at each epoch point
_new_events = Events(events=events[0:-1:int(epoch_length * assr_frequency)])
reader.output_node.events = _new_events

# %%
# Start the pipeline
# =============================
# Some processing to obtain clean epochs


pipeline = PipePool()
pipeline['referenced'] = ReferenceData(reader, reference_channels=['Cz'], invert_polarity=True)
pipeline['channel_cleaned'] = AutoRemoveBadChannels(pipeline['referenced'])
pipeline['down_sampled'] = ReSampling(pipeline['channel_cleaned'], new_sampling_rate=200 * u.Hz)
pipeline['time_filtered_data'] = FilterData(pipeline['down_sampled'],
                                            high_pass=2.0 * u.Hz,
                                            low_pass=90.0 * u.Hz)
pipeline['time_epochs'] = EpochData(pipeline['time_filtered_data'],
                                    event_code=1.0,
                                    base_line_correction=False)
pipeline['dss_time_epochs'] = CreateAndApplySpatialFilter(
    pipeline['time_epochs'],
    components_to_plot=np.arange(0, 5),
    projection_domain=Domain.frequency,
    test_frequencies=test_frequencies,
    return_figures=True)
pipeline.run()

# %%
# Append GFP channel
# ---------------------------------------
# Now we keep events at intervals that correspond to our desired epoch length

pipeline['time_ave_dss'] = AverageEpochs(pipeline['dss_time_epochs'])
pipeline['time_ave'] = AverageEpochs(pipeline['time_epochs'])
pipeline.run()
##
pipeline['time_ave_dss_with_gfp'] = AppendGFPChannel(pipeline['time_ave_dss'])
pipeline['time_ave_with_gfp'] = AppendGFPChannel(pipeline['time_ave'])

pipeline.run()

# plot time frequency response
pipeline['time_freq_transformation'] = TimeFrequencyResponse(pipeline['time_ave_dss_with_gfp'])
pipeline['plot_time_freq'] = PlotTimeFrequencyData(pipeline['time_freq_transformation'],
                                                   plot_x_lim=[0, 1],
                                                   topographic_channels=np.array(['P8']),
                                                   db_scale=False)
pipeline.run()
# %%
# Run F-Test
# ---------------------------------------
# Now we run an F-test to determine if frequency component is significant

pipeline['fft_ave_f_tests_dss'] = FTest(pipeline['time_ave_dss_with_gfp'],
                                        n_fft=int(epoch_length * fs * 1),
                                        test_frequencies=np.append(random_frequencies,
                                                                   [assr_frequency, 2 * assr_frequency]),
                                        delta_frequency=15.0 * u.Hz)

pipeline['fft_ave_f_tests'] = FTest(pipeline['time_ave_with_gfp'],
                                    n_fft=int(epoch_length * fs * 1),
                                    test_frequencies=np.append(random_frequencies,
                                                               [assr_frequency, 2 * assr_frequency]),
                                    delta_frequency=9.0 * u.Hz)
pipeline.run()
pipeline['fft_ave_f_tests_dss'].output_node.statistical_tests['f_test_freq'][
    ["test_name", "df_1", "df_2", "f", "f_critic", "p_value"]].head()

# %%
# Generate figures
# ---------------------------------------
# Now plot the results


pipeline['topographic_map_1'] = PlotTopographicMap(pipeline['fft_ave_f_tests_dss'],
                                                   topographic_channels=np.array(['C4', 'CP2', 'GFP']),
                                                   plot_x_lim=[0, 90],
                                                   plot_y_lim=[0, 3],
                                                   return_figures=True,
                                                   user_naming_rule='dss')
pipeline['topographic_map_2'] = PlotTopographicMap(pipeline['fft_ave_f_tests'],
                                                   topographic_channels=np.array(['C4', 'CP2', 'GFP']),
                                                   plot_x_lim=[0, 90],
                                                   plot_y_lim=[0, 3],
                                                   return_figures=True,
                                                   user_naming_rule='standard')
pipeline['plotter'] = PlotWaveforms(pipeline['fft_ave_f_tests'],
                                    ch_to_plot=np.array(['C4', 'CP2', 'GFP']),
                                    overlay=[pipeline['fft_ave_f_tests_dss']],
                                    plot_x_lim=[0, 90],
                                    plot_y_lim=[-1, 1.5],
                                    offset_step=0.3 * u.uV,
                                    statistical_test='f_test_freq',
                                    show_following_stats=['frequency_tested', 'f'],
                                    user_naming_rule='waveforms',
                                    return_figures=True,
                                    fig_format='.png')
pipeline.run()

# %%
# Save results to a database
# ---------------------------------------
# We get the measurements we are interested in and save them into a database
subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')
_parameters = {'Type': 'ASSR'}
database_path = reader.input_node.paths.test_path.joinpath('assr_f_test_data.sqlite')
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['fft_ave_f_tests'],
                                                      pipeline['fft_ave_f_tests_dss']],
                                      include_waveforms=True
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)

# %%
# Read frequency amplitude from generated database and plot the topographic map
# ------------------------------------_______________________________________________
df = sqlite_tables_to_pandas(database_path=database_path,
                             tables=['peaks_frequency'])

fig_out = plot_topographic_maps(dataframe=df.peaks_frequency,
                                channels_column='channel',
                                cols_by='x',
                                topographic_value='amplitude',
                                layout='biosemi32.lay',
                                title_by='col',
                                title_v_offset=-0.05,
                                color_map_label='Amplitude [$\mu$V]'
                                )



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from scipy import interpolate
import gc
# from pykCSD.pykCSD import KCSD

__author__ = 'jaime undurraga'


def interpolate_potential_fields(x, y, z, **kwargs):
    grid = kwargs.get('grid', 1000j)
    max_distance = np.max(np.linalg.norm((x, y), axis=0))
    grid_x, grid_y = np.mgrid[-max_distance:max_distance:grid, -max_distance:max_distance:grid]
    mask_int = np.array([])
    if x.size > 1 and y.size > 1 and z.size > 1:
        try:
            int_data = interpolate.griddata(points=np.hstack((x, y)), values=np.squeeze(z), xi=(grid_x, grid_y),
                                            method='cubic', rescale=True)
        except ValueError:
            int_data = interpolate.griddata(points=np.hstack((x, y)), values=np.squeeze(z), xi=(grid_x, grid_y),
                                            method='nearest', rescale=True)
        int_data = extrapolate_nans(grid_x, grid_y, int_data)

        radius = int_data.shape[0] / 2.0
        mask_int = mask_data(int_data, int_data.shape[0]/2, int_data.shape[1]/2, radius)

    return mask_int, np.maximum(grid_x.max(), grid_y.max())


def extrapolate_nans(x, y, v):
    '''
    Extrapolate the NaNs or masked values in a grid INPLACE using nearest
    value.

    .. warning:: Replaces the NaN or masked values of the original array!

    Parameters:

    * x, y : 1D arrays
        Arrays with the x and y coordinates of the data points.
    * v : 1D array
        Array with the scalar value assigned to the data points.

    Returns:

    * v : 1D array
        The array with NaNs or masked values extrapolated.
    '''

    if np.ma.is_masked(v):
        nans = v.mask
    else:
        nans = np.isnan(v)
    notnans = np.logical_not(nans)
    v[nans] = interpolate.griddata((x[notnans], y[notnans]), v[notnans],
                                   (x[nans], y[nans]), method='nearest').ravel()
    return v


def mask_data(data, center_x, center_y, radius):
    # mask data from center of the circle
    ny, nx = data.shape
    ix, iy = np.meshgrid(np.arange(nx), np.arange(ny))
    distance = np.sqrt((ix - center_x)**2 + (iy - center_y)**2)
# Mask portions of the data array outside of the circle
    data = np.ma.masked_where(distance > radius, data)
    return data


def plot_clipped(data, center_x, center_y, radius):
    """Plots the image clipped outside of a circle by using a clip path"""
    fig = plt.figure()
    ax = fig.add_subplot(111)

    # Make a circle
    circ = patches.Circle((center_x, center_y), radius, facecolor='none')
    ax.add_patch(circ)  # Plot the outline

    # Plot the clipped image
    ax.imshow(data, clip_path=circ, clip_on=True)
    ax.title('Clipped Array')
    plt.close(fig)
    gc.collect()


def get_3D_spherical_positions(x, y):
    r = np.max(np.concatenate((x, y)))
    pos = [(_x, _y, (r ** 2.0 - (_x ** 2.0 + _y ** 2.0)) ** 0.5) for _x, _y in zip(x, y)]
    return np.squeeze(np.array(pos))



import matplotlib.pyplot as plt
from peegy.io.eeg.reader import eeg_reader
from peegy.io.exporters.edf_bdf_writer import data_to_bdf
from peegy.io.external_tools.file_tools import get_files_and_meta_files
import numpy as np
from peegy.processing.events.event_tools import get_events, events_to_samples_array
import os
from pathlib import Path

_path = Path(os.path.abspath(os.path.dirname(__file__)))
data_folder = _path.parent.parent.absolute().joinpath("test_data/")
data_folder = r'/home/jundurraga/Documents/Measurements/ITD-FR-Pilots/JU/'
data_files_meta = get_files_and_meta_files(data_folder, file_types=['bdf'])
_first_file = data_files_meta.iloc[0].data_links.data_file
data_1 = eeg_reader(file_name=_first_file)
header = data_1._header
data_ori, events_ori, units_ori, annotations_ori = data_1.get_data()
et = get_events(event_channel=events_ori, fs=data_1.fs)
events_array = events_to_samples_array(events=et, fs=data_1.fs, n_samples=data_1.shape[0])

data_to_bdf(output_file_name=_path.parent.parent.absolute().joinpath('deleteme_resampled.bdf'),
            data=data_ori,
            events=events_array,
            header=header,
            fs=data_1.fs)

data_2 = eeg_reader(file_name=_path.parent.parent.absolute().joinpath('deleteme_resampled.bdf'))
data_copy, events_copy, _, _ = data_2.get_data()
et_copy = get_events(event_channel=events_copy, fs=data_2.fs)
et_copy.summary()
fig1 = plt.figure(figsize=(7, 6), dpi=100)
t_original = np.arange(data_ori.shape[0]) / data_1.fs
t_copy = np.arange(data_copy.shape[0]) / data_2.fs
ax = fig1.add_subplot(1, 1, 1)
ax.plot(t_original, data_ori[:, 0])
ax.plot(t_original, events_ori)
ax.plot(t_copy, data_copy[:, 0])
ax.plot(t_copy, events_copy)
plt.show()



import numpy as np
from scipy import interpolate
from scipy.signal import hilbert
from peegy.processing.tools.multiprocessing.multiprocessesing_filter import filt_data
from peegy.processing.tools.filters.eegFiltering import bandpass_fir_win
import astropy.units as u
from peegy.tools.units.unit_tools import set_default_unit
import os
import json


def aep(fs=16384.0 * u.Hz,
        time_length=1.0 * u.s):
    """
    This function generates a typical auditory cortex waveform (as shown in  Human Auditory Evoked Potentials
    Terence W. Picton
    :param fs: sampling rate of the generated waveform (samples per seconds)
    :param time_length: desired time of the output
    :return: waveform and time vector
    """
    fs = set_default_unit(fs, u.Hz)
    time_length = set_default_unit(time_length, u.s)
    time = np.arange(0, time_length * fs) / fs
    # Terrace Picton peaks latencies (in ms)
    _peak_times = np.array([0.0,
                            0.5,
                            0.75,
                            1.5, 2.0,  # I
                            2.5, 3.0,  # II
                            3.5, 4.0,  # III
                            4.5, 5.0,  # IV
                            6.0, 7.0,  # V
                            7.5, 9.0,  # VI-N0
                            13.0, 18.0,  # P0-Na
                            30.0, 40.0,  # Pa-Nb
                            50., 100,  # P1-N1
                            180.0, 300.,   # P2-N2
                            600.0,
                            1000.0
                            ])*1e-3

    # Amplitudes in uV
    _peak_amplitudes = np.array([0.0,
                                 0.0,
                                 0.0,
                                 0.15, 0.0,  # I
                                 0.1, -0.02,  # II
                                 0.2, 0.03,  # III
                                 0.4, 0.35,  # IV
                                 0.5, 0.1,  # V
                                 0.15, -0.4,  # VI-N0
                                 0.1, -0.7,  # P0-Na
                                 0.8, -0.7,  # Pa-Nb
                                 0.9, -2.0,  # P1-N1
                                 2.5, -1.0,  # P2-N2
                                 -0.2,
                                 0
                                 ])
    _peak_amplitudes = _peak_amplitudes - np.mean(_peak_amplitudes, axis=0)
    f = interpolate.interp1d(_peak_times, _peak_amplitudes, kind='linear', fill_value="extrapolate")
    _x = np.append(0, np.logspace(np.log10(time[1].value), np.log10(time[-1].value), num=75, endpoint=False))
    _y = f(_x)
    f2 = interpolate.interp1d(_x, _y, kind='quadratic', fill_value="extrapolate")
    ynew = f2(time).reshape(-1, 1) * u.uV
    # plt.plot(_x, _y)
    # plt.plot(_peak_times, _peak_amplitudes, 'o')
    # plt.plot(time, ynew)
    # plt.show()
    w = np.zeros(ynew.shape)
    _n_samples = np.minimum(int(0.3 * u.s * fs), ynew.shape[0])
    w_template = fade_in_out_window(n_samples=_n_samples, fraction=0.2)
    w[0:w_template.shape[0]] = w_template
    ynew = ynew * w
    return ynew, time


def eog(fs=16384.0 * u.Hz,
        thau1: u.quantity.Quantity = 0.05 * u.s,
        thau2: u.quantity.Quantity = 0.01 * u.s,
        peak_amp: u.quantity.Quantity = 100 * u.uV,
        event_duration: u.quantity.Quantity = 1.0 * u.s):
    """
    This function generates a typical eye blink waveform with an amplitude of 300 uV
    :param fs: sampling rate of the generated waveform (samples per seconds)
    :return: waveform and time vector
    """
    fs = set_default_unit(fs, u.Hz)
    thau1 = set_default_unit(thau1, u.s)
    thau2 = set_default_unit(thau2, u.s)
    peak_amp = set_default_unit(peak_amp, u.uV)
    event_duration = set_default_unit(event_duration, u.s)
    event_time = np.arange(0, event_duration * fs) / fs
    # event = windows.flattop(event_time.size) + windows.general_gaussian(event_time.size, p=1.5, sig=5000)
    event = (np.exp(-event_time/thau1) - np.exp(-event_time/thau2))
    y_new = np.zeros(event_time.shape)
    y_new[0:event_time.size] = event
    # demean
    y_new = y_new - np.mean(y_new)
    # apply rise-fade window
    fade_in_out = np.ones(y_new.shape)
    fade_len = np.round(event_time.size // 20)
    fade_in_out[0: fade_len] = (np.sin(2 * np.pi * (1 / (4 * fade_len / fs)) *
                                       event_time[0: fade_len] * u.rad) ** 2)
    fade_in_out[-fade_len:] = (np.sin(2 * np.pi * (1 / (4 * fade_len / fs)) *
                                      event_time[0: fade_len] * u.rad + np.pi * u.rad / 2) ** 2)
    y_new = y_new * fade_in_out * u.uV
    y_new *= peak_amp / np.abs(y_new).max()

    return y_new.reshape(-1, 1), event_time


def eog_template(fs=16384.0 * u.Hz,
                 event_duration: u.quantity.Quantity = 1.4 * u.s,
                 fade_duration: u.Quantity = 0.2 * u.s):
    """
    This function reads a eog template and resample it to the target sampling rate via interpolation.
    See Valderrama, J. T., de la Torre, A., & Van Dun, B. (2018). An automatic algorithm for blink-artifact suppression
    based on iterative template matching: Application to single channel recording of cortical auditory evoked
    potentials. Journal of Neural Engineering, 15(1), 016008. https://doi.org/10.1088/1741-2552/aa8d95
    :param fs: sampling rate of the generated waveform (samples per seconds)
    :return: waveform and time vector
    """
    fs = set_default_unit(fs, u.Hz)
    event_duration = set_default_unit(event_duration, u.s)
    time = np.arange(0, event_duration * fs) / fs
    _path = os.path.abspath(__file__)
    pathname = os.path.dirname(_path)
    with open(pathname + os.sep + 'h0.json') as _f:
        data = json.load(_f)
    units = u.Quantity(1, data['units'])
    template = np.array(data['eog_template']) * units
    fs_data = data['fs']
    time_data = np.arange(0, template.shape[0]) / fs_data
    f = interpolate.interp1d(time_data, template.squeeze(), kind='linear', fill_value="extrapolate")
    y_new = f(time).reshape(-1, 1) * units
    _window = np.ones((y_new.size, 1))
    nw = int(np.minimum(np.maximum(fade_duration * fs, 0), template.shape[0] // 2))
    _window[0: nw] = (np.arange(0, nw) / (nw - 1)).reshape(-1, 1)
    _window[_window.size - nw::] = _window[0: nw][::-1]
    y_new = y_new * _window
    return y_new, event_duration


def fade_in_out_window(n_samples: int | None = None, fraction=0.1):
    """
    This function will generate a fade window with a fade_in and fade_out length as a propoertion of the number of
    samples
    :param n_samples: length of the window
    :param fraction: percentage of fade in and out length
    :return:
    """
    # apply rise-fade window
    fade_in_out = np.ones([n_samples, 1])
    fade_len = int(np.round(n_samples * fraction))
    fade_in_out[0: fade_len, 0] = (np.sin(2 * np.pi * (np.arange(fade_len) * u.rad / (4 * fade_len))) ** 2)
    fade_in_out[-fade_len:, 0] = (np.sin(2 * np.pi * (np.arange(fade_len) * u.rad / (4 * fade_len))
                                         + np.pi * u.rad / 2) ** 2)
    return fade_in_out


def artifact_and_brain_envelope(fs: u.Quantity = 16384.0 * u.Hz,  # sampling rate
                                stimulus_delay: u.Quantity = 0.0 * u.s,  # brainstem delay
                                brain_delay: u.Quantity = 113 * u.ms,  # brainstem delay
                                duration: u.Quantity = 1 * u.s,  # seconds
                                seed=None,  # random generator seed
                                leak_amplitude: u.Quantity = 20.0 * u.uV,
                                brain_amplitude: u.Quantity = 0.2 * u.uV
                                ):
    """
    This function will generate an envelope response (as it was from the brain) evoked by an external stimulus.
    :param fs: sampling rate
    :param stimulus_delay: physical delay between the presented waveform and the output of the device (this simulates
    a delay between the waveform and the transducer)
    :param brain_delay: delay of the envelope response being followed by the brain (this delay emulates a neural source
    responding later in time to the stimulus)
    :param duration: length of the envelope response
    :param seed: defines the seed for the noise generator (this allows to reproduce the same stimulus every time)
    :param leak_amplitude: the amplitude of the leaked waveform
    :param brain_amplitude: the amplitude of the brain response
    :return: brain_waveform (the target neural response),
    stimulus_waveform (the presented waveform),
    leaked_stimulus (the recorded artefact in an EEG system; this is, the stimulus_waveform with delay and filter
    applied)
    """
    fs = set_default_unit(fs, u.Hz)
    duration = set_default_unit(duration, u.s)
    stimulus_delay = set_default_unit(stimulus_delay, u.s)
    brain_delay = set_default_unit(brain_delay, u.s)
    leak_amplitude = set_default_unit(leak_amplitude, u.uV)
    brain_amplitude = set_default_unit(brain_amplitude, u.uV)

    time = np.arange(0, duration.to(u.s).value, 1 / fs.to(u.Hz).value).reshape(-1, 1) * u.s

    # define stimulation waveform
    if seed is not None:
        np.random.seed(seed)
    stimulus_waveform = np.random.randn(time.size, 1) * u.dimensionless_unscaled
    stimulus_waveform = stimulus_waveform / np.abs(stimulus_waveform).max()

    # apply rise-fade window
    fade_in_out = np.ones(stimulus_waveform.shape)
    fade_len = round(time.size // 20)
    fade_in_out[0: fade_len] = np.sin(2 * np.pi * u.rad * (1 / (4 * fade_len / fs)) * time[0: fade_len]) ** 2
    fade_in_out[-fade_len:] = np.sin(2 * np.pi * u.rad * (1 / (4 * fade_len / fs)) *
                                     time[0: fade_len] + np.pi * u.rad / 2) ** 2
    stimulus_waveform = stimulus_waveform * fade_in_out

    # brain response follows signal envelop with a delay
    template_waveform = np.abs(hilbert(stimulus_waveform)) * stimulus_waveform.unit
    # rectify  brain response
    template_waveform[template_waveform < 0] = 0

    # filter brain response
    _b_brain = bandpass_fir_win(high_pass=2 * u.Hz,
                                low_pass=80.0 * u.Hz,
                                fs=fs)

    template_waveform = filt_data(data=template_waveform, b=_b_brain)
    template_waveform = template_waveform * fade_in_out

    # pad zeros to add stimulus and neural delay
    template_waveform = np.pad(template_waveform, ((int(fs * stimulus_delay) + int(fs * brain_delay), 0), (0, 0)),
                               'constant', constant_values=(0, 0))
    template_waveform = template_waveform / np.abs(template_waveform).max()

    # apply delay to leaked stimulus (assuming system delay)
    leaked_stimulus = np.pad(stimulus_waveform, ((int(fs * stimulus_delay), int(fs * brain_delay)),
                                                 (0, 0)), 'constant', constant_values=(0, 0))
    # pad with zeros stimulus_waveform to output all data with same length
    stimulus_waveform = np.pad(stimulus_waveform, ((0, int(fs * stimulus_delay) + int(fs * brain_delay)),
                                                   (0, 0)), 'constant', constant_values=(0, 0))

    # leaked artifact
    _b = bandpass_fir_win(None, 1000 * u.Hz, fs=fs, ripple_db=60)
    leaked_stimulus = filt_data(data=leaked_stimulus, b=_b)

    # scale respective signals
    brain_waveform = template_waveform * brain_amplitude
    leaked_stimulus = leaked_stimulus * leak_amplitude
    return brain_waveform, stimulus_waveform, leaked_stimulus


def gaussian(x, mu, sig):
    return np.exp(-np.power(x - mu, 2.) / (2 * np.power(sig, 2.)))


def abr(fs: u.quantity.Quantity = 16384.0 * u.Hz,
        time_length: u.quantity.Quantity = 1.0 * u.s,
        peaks_amplitudes: np.ndarray = np.array([0.1, 0.07, 0.2, 0.05, 0.5]) * u.uV):
    """
    This function generates a typical abr waveform
    :param fs: sampling rate of the generated waveform (samples per seconds)
    :param time_length: desired time of the output
    :param peaks_amplitudes: array with amplitudes for waves i, ii, iii, iv, and v
    :return: waveform and time vector
    """
    fs = set_default_unit(fs, u.Hz)
    time_length = set_default_unit(time_length, u.s)
    peaks_amplitudes = set_default_unit(peaks_amplitudes, u.uV)

    time = np.arange(0, time_length * fs) / fs
    _template_time = np.arange(0, 0.01 * u.s * fs) / fs
    wave_i = np.concatenate(([0], np.diff(gaussian(_template_time, 0.002 * u.s, 0.0001 * u.s)))) + 0.1 * gaussian(
        _template_time, 0.002 * u.s, 0.0001 * u.s)
    wave_ii = np.concatenate(([0], np.diff(gaussian(_template_time, 0.003 * u.s, 0.00015 * u.s)))) + 0.1 * gaussian(
        _template_time, 0.003 * u.s, 0.0001 * u.s)
    wave_iii = 2 * np.concatenate(([0], np.diff(gaussian(_template_time, 0.004 * u.s, 0.00015 * u.s)))) + 1 * gaussian(
        _template_time, 0.004 * u.s, 0.0001 * u.s)
    wave_iv = np.concatenate(([0], np.diff(gaussian(_template_time, 0.005 * u.s, 0.0001 * u.s)))) + gaussian(
        _template_time, 0.005 * u.s, 0.0001 * u.s)
    wave_v = 2 * np.concatenate(([0], np.diff(gaussian(_template_time, 0.006 * u.s, 0.001 * u.s)))) + .5 * gaussian(
        _template_time, 0.006 * u.s, 0.00038 * u.s)
    # normalize and scale peaks
    wave_i *= peaks_amplitudes[0] / wave_i.max()
    wave_ii *= peaks_amplitudes[1] / wave_ii.max()
    wave_iii *= peaks_amplitudes[2] / wave_iii.max()
    wave_iv *= peaks_amplitudes[3] / wave_iv.max()
    wave_v *= peaks_amplitudes[4] / wave_v.max()
    _abr = wave_i + wave_ii + wave_iii + wave_iv + wave_v
    f = interpolate.interp1d(_template_time, _abr, kind='linear', fill_value="extrapolate")
    ynew = f(time).reshape(-1, 1) * u.uV
    w = np.zeros(ynew.shape)
    _n_samples = np.minimum(_template_time.shape[0], ynew.shape[0])
    w_template = fade_in_out_window(n_samples=_n_samples, fraction=0.2)
    w[0:w_template.shape[0]] = w_template
    ynew = ynew * w

    return ynew, time



import numpy as np
import multiprocessing
import pyfftw
import psutil as psu
import scipy.signal as signal
import time
import ctypes
import astropy.units as u
from tqdm import tqdm
__author__ = 'jundurraga-ucl'


def timeit(method):
    def timed(*args, **kw):
        time_start = time.time()
        result = method(*args, **kw)
        time_end = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((time_end - time_start) * 1000)
        else:
            print('{:s}: {:.2f} ms'.format(method.__name__, (time_end - time_start) * 1000))
        return result
    return timed


def find_freq(freq_vector=np.array, freq_to_find=np.array, f_range=0.0):
    f_bins = []
    for f in freq_to_find:
        f_pos = [np.where(freq_vector < f - f_range)[0][-1],
                 np.where(freq_vector > f + f_range)[0][0]]
        f_bins.append(f_pos)
    return f_bins


def eeg_notch_filter(x: np.array = np.array([]), f: list = [50.0], f_range=1.0, fs=16384.0, blocks=8) -> np.array:
    if np.mod(x.shape[0], 2) == 0:
        fft_size = (x.shape[0] / 2) + 1
    else:
        fft_size = (x.shape[0] + 1) / 2

    freq = np.arange(fft_size) * fs / x.shape[0]
    f_bins = find_freq(freq, f, f_range)
    blocks = np.minimum(x.shape[1], blocks)
    xfilt = np.zeros(x.shape)
    for b in np.arange((np.ceil(x.shape[1]) / blocks)):
        pos_ini = int(b * blocks)
        pos_end = int(np.minimum((b + 1) * blocks, x.shape[1]) + 1)
        fft = pyfftw.builders.rfft(x[:, pos_ini:pos_end], overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
                                   threads=multiprocessing.cpu_count())
        spectrum = fft()
        for i, _f_pos in enumerate(f_bins):
            spectrum[_f_pos[0]:_f_pos[1] + 1, :] = 0.0

        ifft = pyfftw.builders.irfft(spectrum, overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
                                     threads=multiprocessing.cpu_count())
        xfilt[:, pos_ini:pos_end] = ifft()
        print('fft notch filtered:', pos_end - 1)
        if pos_end == x.shape[1]:
            break
    return xfilt


@timeit
def filt_filt_multithread(input_data: np.array([]),
                          a=None,
                          b=None,
                          ch_per_block=None,
                          n_jobs=None) -> np.array:
    mp_arr = multiprocessing.Array(ctypes.c_double, input_data.size, lock=False)
    arr = np.frombuffer(mp_arr)
    data = arr.reshape(input_data.shape)
    np.copyto(data, input_data)

    if not n_jobs:
        n_jobs = np.maximum(multiprocessing.cpu_count(), 1)
    else:
        n_jobs = np.minimum(multiprocessing.cpu_count(), n_jobs)
    if not ch_per_block:
        _virtual_memory = psu.virtual_memory()
        ch_per_block = np.maximum(1, int(_virtual_memory.available /
                                         (data.nbytes / data.shape[1]) * 0.1))
    ch_per_block = np.minimum(data.shape[1], ch_per_block)
    print(('filtering ' + str(ch_per_block) + ' channels per block'))

    for _bl in np.arange((np.ceil(data.shape[1]) / ch_per_block)):
        p = []
        pos_ini = int(_bl * ch_per_block)
        pos_end = int(np.minimum((_bl + 1) * ch_per_block, data.shape[1]))
        _b_size = pos_end - pos_ini
        _ch_per_thread = int(np.ceil(float(_b_size) / n_jobs))
        _n_jobs = int(np.minimum(n_jobs, np.ceil(float(_b_size) / _ch_per_thread)))
        # q_out = multiprocessing.Queue(maxsize=_n_jobs)
        pool = multiprocessing.Pool(processes=_n_jobs)
        for _t in range(_n_jobs):
            sub_pos_ini = int(pos_ini + _ch_per_thread * _t)
            sub_pos_end = int(np.minimum(pos_end, pos_ini + _ch_per_thread * (_t + 1)))
            p.append(pool.Process(target=filter_worker,
                                  args=(data, sub_pos_ini,
                                        sub_pos_end, b, a)))
            print(('Thread %i filtering channels: %i to %i' % (_t, sub_pos_ini, sub_pos_end - 1)))
        [_p.start() for _p in p]
        [_p.join() for _p in p]
        pool.close()
        pool.join()
        print(('Filtered channels: %i to %i' % (pos_ini, pos_end - 1)))
    return data


def filter_worker(data, pos_ini, pos_end, b, a):
    """
    IIR/FIR filtering worker. This function filter part of input data. Filter coefficients can be either  FIR or IIR
    :param data: mxn matrix to be filtered
    :param pos_ini: initial column to begin filtering
    :param pos_end: end column to begin filtering
    :param b: filter coefficients numerator
    :param a: filter coefficients denominator
    :return: partially filtered data
    :return: filtered data
    """
    data[:, pos_ini:pos_end] = signal.filtfilt(b, a, data[:, pos_ini:pos_end], axis=0)


@timeit
def filt_filt_multithread_ovs(input_data: np.array([]),
                              b=None,
                              ch_per_block=None,
                              n_jobs=None,
                              onset_padding=True) -> np.array:
    """
    This function implements parallel filtering using overlap-saving FIR filtering
    :param input_data: numpy array with data to be filtered
    :param b: FIR filter coefficients
    :param ch_per_block: number of channels to be filter per job
    :param n_jobs: number of parallel jobs
    :param onset_padding: whether to pad samples or not at the beginning and end of the input data. This minimizes the
    ringing effects of the filter
    :return: filtered data
    """
    mp_arr = multiprocessing.Array(ctypes.c_double, input_data.size, lock=False)
    arr = np.frombuffer(mp_arr)
    pad_width = 0
    if onset_padding:
        pad_width = input_data.shape[0] // 8
        input_data = np.pad(input_data, ((pad_width, pad_width), (0, 0)), mode='edge')
    data = arr.reshape(input_data.shape)
    np.copyto(data, input_data)

    if not n_jobs:
        n_jobs = np.maximum(multiprocessing.cpu_count(), 1)
    else:
        n_jobs = np.minimum(multiprocessing.cpu_count(), n_jobs)
    if not ch_per_block:
        _virtual_memory = psu.virtual_memory()
        ch_per_block = np.maximum(1, int(_virtual_memory.available /
                                         (data.nbytes / data.shape[1]) * 0.1))
    ch_per_block = np.minimum(data.shape[1], ch_per_block)
    print(('filtering ' + str(ch_per_block) + ' channels per block'))

    for _bl in tqdm(np.arange((np.ceil(data.shape[1]) / ch_per_block)), desc='Filter'):
        p = []
        pos_ini = int(_bl * ch_per_block)
        pos_end = int(np.minimum((_bl + 1) * ch_per_block, data.shape[1]))
        _b_size = pos_end - pos_ini
        _ch_per_thread = int(np.ceil(float(_b_size) / n_jobs))
        _n_jobs = int(np.minimum(n_jobs, np.ceil(float(_b_size) / _ch_per_thread)))
        # q_out = multiprocessing.Queue(maxsize=_n_jobs)
        pool = multiprocessing.Pool(processes=_n_jobs)
        for _t in range(_n_jobs):
            sub_pos_ini = int(pos_ini + _ch_per_thread * _t)
            sub_pos_end = int(np.minimum(pos_end, pos_ini + _ch_per_thread * (_t + 1)))
            p.append(pool.Process(target=filt_worker_ovs,
                                  args=(data, sub_pos_ini,
                                        sub_pos_end, b)))
            # print(('Thread %i filtering channels: %i to %i' % (_t, sub_pos_ini, sub_pos_end - 1)))
        [_p.start() for _p in p]
        [_p.join() for _p in p]
        pool.close()
        pool.join()
        # print(('Filtered channels: %i to %i' % (pos_ini, pos_end - 1)))
    return data[pad_width: data.shape[0] - pad_width, :]


def filt_worker_ovs(data: type(np.array) | None = None, pos_ini=0, pos_end=0, b: type(np.array) | None = None):
    """
    Overlap saving filtering worker. This function filter part of input data. Filter coefficients are assumed to be
    from an FIR filter
    :param data: mxn matrix to be filtered
    :param pos_ini: initial column to begin filtering
    :param pos_end: end column to begin filtering
    :param b: filter coefficients
    :return: partially filtered data
    """
    if pos_end is None:
        pos_end = data.shape[1]
    s_1 = ols_filt(b=b, x=data[:, pos_ini:pos_end])
    data[:, pos_ini:pos_end] = s_1[0:data.shape[0], :]
    return data


def filt_filt_worker_ovs(data: type(np.array) | None = None, pos_ini=0, pos_end=None, b: type(np.array) | None = None):
    """
    Overlap saving filtering worker. This function filter part of input data. Filter coefficients are assumed to be
    from an FIR filter
    :param data: mxn matrix to be filtered
    :param pos_ini: initial column to begin filtering
    :param pos_end: end column to begin filtering
    :param b: filter coefficients
    :return: partially filtered data
    """
    if pos_end is None:
        pos_end = data.shape[1]
    data[:, pos_ini:pos_end] = ols_filt_filt(b=b, x=data[:, pos_ini:pos_end])
    return data


def ols_filt(b: type(np.array) | None = None, x: type(np.array) | None = None, axis=0):
    """
    Overlap-saving filtering.
    Filter a one-dimensional array with an FIR filter
    Filter a data array using a FIR filter given in `b`.
    Filtering uses the overlap-add method converting both `x` and `b`
    into frequency domain first.  The FFT size is determined as the
    next higher power of 2 of twice the length of `b`.

    :param b: one-dimensional numpy array. The impulse response of the filter
    :param x: numpy array to be filtered
    :param axis: dimension in which filtering is applied
    :return: filtered array
    """

    l_i = b.shape[0]
    # Find power of 2 larger that 2*l_i (from abarnert on Stackoverflow)
    l_f = 2 << (l_i-1).bit_length()
    l_s = l_f - l_i + 1
    l_sig = x.shape[0]
    offsets = range(0, l_sig, l_s)

    # handle complex or real input
    if np.iscomplexobj(b) or np.iscomplexobj(x):
        fft_func = np.fft.fft
        ifft_func = np.fft.ifft
        res = np.zeros((l_sig+l_f, *x.shape[1:]), dtype=np.complex128) * x.unit
    else:
        fft_func = np.fft.rfft
        ifft_func = np.fft.irfft
        res = np.zeros((l_sig+l_f, *x.shape[1:])) * x.unit

    FDir = np.atleast_2d(fft_func(b, n=l_f, axis=axis)).T
    if x.ndim == 3:
        FDir = np.atleast_3d(FDir)

    # overlap and add
    for n in offsets:
        res[n:n+l_f, :] += ifft_func(fft_func(x[n:n+l_s, :], n=l_f, axis=0)*FDir, axis=0)

    return res


def ols_filt_filt(b: type(np.array) | None = None, x: type(np.array) | None = None, axis=0):
    """
    Overlap-saving filt-filt. Filters input x twice (in both directions) to cancel out the group-delay introduced by the
    filter
    :param b: FIR filter coef
    :param x: numpy array to be filtered
    :param axis: dimension in which filtering is applied
    :return: filtered array
    """
    s_1 = ols_filt(b=b, x=x, axis=axis)
    _l_1 = s_1.shape[0] - x.shape[0]
    s_1 = ols_filt(b=b, x=np.flip(s_1, axis=axis), axis=axis)
    s_1 = np.flip(s_1, axis=axis)
    return s_1[_l_1:_l_1 + x.shape[0], :]


def bandpass_fir_win(high_pass: u.quantity.Quantity | None = None,
                     low_pass: u.quantity.Quantity | None = None,
                     fs: u.quantity.Quantity = 16884.0 * u.Hz,
                     transition_width: u.Quantity = 1.0 * u.Hz,
                     ripple_db=60.0,
                     filt_filt_cutoff=False,
                     n_passes: int = 2
                     ):
    """
    Generate FIR filter coefficients using a Kaiser filter
    :param high_pass: frequency (in Hz) of high-pass filter
    :param low_pass: frequency (in Hz) of low-pass filter
    :param fs: sampling frequency (in Hz)
    :param transition_width: width of filter transition band (in Hz)
    :param ripple_db: amount of ripple, in dB, in the pass-band and rejection regions. The magnitude variation will be
    below -ripple dB within the bandpass region, whilst the attenuation in the rejection band will be at lesast +ripple
    dB
    :param filt_filt_cutoff: if True, the effective cutoff frequencies obtained when applying filtfilt will be shown
    :return: a numpy array with the filter coefficients
    """
    nyq = 0.5 * fs
    width = transition_width / nyq
    taps = None
    if low_pass is not None and high_pass is None:
        width = min((nyq - low_pass) / nyq, width)
        n_taps, beta = signal.kaiserord(ripple_db, width)
        # ensure odd number of taps to prevent filter to crash
        n_taps = n_taps // 2 * 2 + 1

        taps = signal.firwin(n_taps,
                             low_pass.to(u.Hz).value,
                             fs=fs.to(u.Hz).value,
                             pass_zero='lowpass',
                             window=('kaiser', beta), scale=True)

    if low_pass is None and high_pass is not None:
        width = min(high_pass / nyq, width)
        n_taps, beta = signal.kaiserord(ripple_db, width)
        n_taps = n_taps // 2 * 2 + 1
        taps = signal.firwin(n_taps, high_pass.to(u.Hz).value,
                             fs=fs.to(u.Hz).value,
                             pass_zero='highpass',
                             window=('kaiser', beta), scale=True)
    if low_pass is not None and high_pass is not None:
        width_1 = min((nyq - low_pass) / nyq, width)
        width_2 = min(high_pass / nyq, width)
        width = min(width_1, width_2)
        n_taps, beta = signal.kaiserord(ripple_db, width)
        n_taps = n_taps // 2 * 2 + 1
        taps = signal.firwin(n_taps, [high_pass.to(u.Hz).value,
                                      low_pass.to(u.Hz).value],
                             fs=fs.to(u.Hz).value, pass_zero='bandpass',
                             window=('kaiser', beta), scale=True)

    print('FIR Kaiser filter ({:} taps) HP {:} - LP {:}, {:} dB ripple, {:} transition width.'.format(
        taps.size,
        high_pass,
        low_pass,
        ripple_db,
        width * nyq))
    if filt_filt_cutoff:
        _hp, _lp = get_cutoff_fir_filt_filt(b=taps,
                                            fs=fs,
                                            n_passes=n_passes)

        print('Effective cutoff frequencies using filtfilt with {:} passes HP {:} - LP {:}. Order of the filter'
              ' will be {:} times the original'.format(n_passes,
                                                       _hp,
                                                       _lp,
                                                       n_passes))

    return taps


def get_cutoff_fir_filt_filt(
        b: u.quantity.Quantity | None = None,
        fs: u.quantity.Quantity = 16884.0 * u.Hz,
        frequency_resolution: u.Quantity = 0.0001 * u.Hz,
        n_passes: int = 2
        ):
    """
    Provides the cutoff frequencies of a FIR filter LP, HP, or bandpass.
    :param b: numpy array with filter coefficients
    :param fs: sampling frequency (in Hz)
    :param frequency_resolution: float indicating the frequency resolution in the frequency-domain.
    :param n_passes: integer indicating the number of times the filter is to be passed.
    :return: a numpy array with the filter coefficients
    """
    worN = int(0.5 * fs / frequency_resolution)
    w, h = signal.freqz(b, worN=worN)
    freq = (w / np.pi) * fs * 0.5
    magnitude = np.abs(h)
    idx = np.argwhere(np.diff(np.sign(magnitude ** n_passes - 0.5))).flatten()
    diff = np.diff(magnitude)
    high_pass = None
    low_pass = None
    if idx.size == 1:
        if diff[idx[0]] > 0:
            high_pass = freq[idx[0]]
        else:
            low_pass = freq[idx[0]]
    else:
        high_pass = freq[idx[0]]
        low_pass = freq[idx[1]]

    return high_pass, low_pass



"""
.. _tut-acc-standard-weighted-average-per-epochs-pipeline-sim:

##################################################################
ACC Standard analysis with by epochs (Simulated)
##################################################################

In this example we simulate an ACC response and average by epochs.
This is useful if you are interested in test how things improve as you collect more and more epochs
"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#     matplotlib.use('Qt5Agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.detection.definitions import TimeROI
from peegy.processing.tools.template_generator.auditory_waveforms import aep
from peegy.processing.pipe.general import FilterData
from peegy.processing.pipe.epochs import EpochData, AverageEpochs
from peegy.processing.pipe.epoching.epoching import EpochsByStep
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
from peegy.io.storage.data_storage_reading_tools import sqlite_tables_to_pandas
import seaborn as sns
import os
import numpy as np
import astropy.units as u

# %%
# Time regions of interest
# ============================
# TimeROI are defined as time regions where different measures will be performed, e.g. SNR measure or statistical
# measures

roi_windows = np.array(
    [
        TimeROI(
            ini_time=50.0 * u.ms,
            end_time=550.0 * u.ms,
            window_label="analysis_window",
        )
    ]
)

# %%
# Generate some data
# ===================
# First we generate some ACC data

n_channels = 32
fs = 512.0 * u.Hz
template_waveform, _ = aep(fs=fs)
n_events = 100
event_times = np.arange(0, n_events, 1.0) * u.s
snr_db = 10
snr = 10 ** (snr_db / 10) / n_events
reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.ones(n_channels)),
                           layout_file_name='biosemi32.lay',
                           snr=snr,
                           event_times=event_times,
                           event_code=1.0,
                           figures_subset_folder='acc_test_per_epochs',
                           include_eog_events=False,
                           noise_attenuation=3,
                           noise_seed=0,
                           f_noise_low=0 * u.Hz,
                           f_noise_high=fs / 2,
                           f_var_low=2 * u.Hz,
                           f_var_high=30 * u.Hz,
                           neural_ini_time_snr=roi_windows[0].ini_time,
                           neural_end_time_snr=roi_windows[0].end_time,
                           )
reader.run()

# %%
# Start the pipeline
# ============================
# Now we proceed with our basic processing pipeline


pipeline = PipePool()
# %%
# Start the pipeline
# ------------------------------------
pipeline['time_filtered_data'] = FilterData(reader,
                                            high_pass=2.0 * u.Hz,
                                            low_pass=30.0 * u.Hz)

pipeline.run()

# %%
# Get Epochs
# ------------------------------------
# We partition the data into epochs or trials based on the event code used.


pipeline['time_epochs'] = EpochData(pipeline['time_filtered_data'],
                                    event_code=1.0)

# %%
# Start per epochs pipeline and compute average
# ------------------------------------------------

pipeline_epochs = PipePool()
pipeline_epochs['average'] = AverageEpochs(pipeline['time_epochs'],
                                           roi_windows=roi_windows,
                                           weighted_average=True,
                                           weight_across_epochs=False)

# now we save our data to a database
subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'ACC'}

database_path = reader.input_node.paths.test_path.joinpath('acc_test_per_epochs.sqlite')
pipeline_epochs['database'] = SaveToDatabase(database_path=str(database_path),
                                             measurement_information=measurement_info,
                                             subject_information=subject_info,
                                             recording_information={'recording_device': 'dummy_device'},
                                             stimuli_information=_parameters,
                                             processes_list=[pipeline_epochs['average']],
                                             include_waveforms=False,
                                             )

pipeline['per_epochs'] = EpochsByStep(at_each_epoch_block_do=pipeline_epochs,
                                      epochs_step_size=2,
                                      deep_break=False,
                                      max_epochs=1000)

pipeline.run()

# %% Plots some results
df = sqlite_tables_to_pandas(database_path=str(database_path),
                             tables=['f_test_time'])['f_test_time']

fig_out = sns.lineplot(data=df,
                       x='n_epochs',
                       y='rn',
                       hue='channel')

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)



import peegy.io.readers.cnt_reader as cntr
import matplotlib.pyplot as plt
import numpy as np
__author__ = 'jundurraga'

f_name = '/mnt/90720477720463F6/Measurements/Kristin_Uhler_MMN/MMN001HL/anmmnA_I70.cnt'
header = cntr.read_cnt_header(file_name=f_name)
ch = np.array([0, 1])
data, events, *_ = cntr.get_data(header=header, ini_time=0, end_time=None)
_events = events.get_events()
fs = header['fs']
x = np.arange(0, data.shape[0]) / fs
fig = plt.figure()
plt.plot(x, data - np.mean(data, axis=0))
plt.plot([_ev.time_pos.value for _ev in _events], [0 for _ev in _events], 'ro')
plt.show(block=True)



import numpy as np
import psutil as psu
import astropy.units as u
from peegy.processing.system.progress import ParallelTqdm
from .. filters.eegFiltering import filt_filt_worker_ovs, filt_worker_ovs
import joblib


def filt_data(data: type(np.array) | None = None,
              b: type(np.array) | None = None,
              n_jobs=None,
              ch_per_block=None,
              onset_padding=True,
              mode='original',
              max_percentage_memory: float = 0.8):
    """
    This function will filter input data. Coefficients must arise from a FIR filter
    :param data: a mxn matrix, data is filter on their first dimension
    :param b: filter coefficients
    :param n_jobs: number of parallel tasks
    :param ch_per_block: number of columns to be filtered by each job
    :param onset_padding: whether to pad data at the beginning and the end of the first dimension. This minimizes
    the ringing effect of the filter
    :param mode: any of these {'original', 'full', 'valid'}. It determines the size of the convolved output
    :param max_percentage_memory: memory limit in percentage used to estimate parallel processing resources
    :return: filtered data (mxn numpy array)
    """
    b = b * u.dimensionless_unscaled
    pad_width = 0
    if onset_padding:
        # we do this to minimize ringing of filter
        pad_width = data.shape[0] // 8
        data = np.vstack((data[1:pad_width + 1, :][::-1], data, data[::-1, :][1:pad_width + 1]))
    filtered_data = data
    if not n_jobs:
        n_jobs = np.maximum(joblib.cpu_count(), 1)
    else:
        n_jobs = np.minimum(joblib.cpu_count(), n_jobs)

    padding_size = ((0, b.size - 1), *((0, 0),) * (filtered_data.ndim - 1))
    padding_value = (((0, 0),) * filtered_data.ndim)
    filtered_data = np.pad(filtered_data, padding_size,
                           'constant',
                           constant_values=padding_value)

    if not ch_per_block:
        _virtual_memory = psu.virtual_memory()
        ch_per_block = np.maximum(1, int(_virtual_memory.available * max_percentage_memory /
                                         (n_jobs * filtered_data.nbytes / filtered_data.shape[1])))
    ch_per_block = np.minimum(filtered_data.shape[1], ch_per_block)
    # print('filtering {:} channels per block'.format(ch_per_block))

    if mode == 'full':
        samples_delay = 0
        samples_delay_end = 0
    if mode == 'original':
        # compensates for delay assuming that the FIR filter coefficients are symmetric
        samples_delay = (b.size - 1) // 2
        samples_delay_end = samples_delay
    if mode == 'valid':
        samples_delay = 0
        samples_delay_end = b.size - 1

    for _bl in np.arange((np.ceil(filtered_data.shape[1]) / ch_per_block)):
        pos_ini = int(_bl * ch_per_block)
        pos_end = int(np.minimum((_bl + 1) * ch_per_block, filtered_data.shape[1]))
        _b_size = pos_end - pos_ini
        _ch_per_thread = int(np.ceil(float(_b_size) / n_jobs))
        _n_jobs = int(np.minimum(n_jobs, np.ceil(float(_b_size) / _ch_per_thread)))
        sub_pos_ini = []
        sub_pos_end = []
        for _t in range(_n_jobs):
            sub_pos_ini.append(int(pos_ini + _ch_per_thread * _t))
            sub_pos_end.append(int(np.minimum(pos_end, pos_ini + _ch_per_thread * (_t + 1))))

        ParallelTqdm(total=_n_jobs,
                     desc='Parallel filtering/worker',
                     n_jobs=_n_jobs,
                     mmap_mode='w+',
                     backend='threading')(
            joblib.delayed(filt_worker_ovs)(filtered_data, sub_pos_ini[_i], sub_pos_end[_i], b) for
            _i in range(_n_jobs))
        # print('Filtered channels: {:} to {:}'.format(pos_ini, pos_end - 1))

    return filtered_data[pad_width + samples_delay:
                         filtered_data.shape[0] - pad_width - samples_delay_end, :]


def filt_filt_data(data: type(np.array) | None = None,
                   b: type(np.array) | None = None,
                   n_jobs=None,
                   ch_per_block=None,
                   onset_padding=True):
    """
    This function will filter input data using a zero group-delay technique. Coefficients must arise from a FIR filter
    :param data: a mxn matrix, data is filter on their first dimension
    :param b: filter coefficients
    :param n_jobs: number of parallel tasks
    :param ch_per_block: number of columns to be filtered by each job
    :param onset_padding: whether to pad data at the beginning and the end of the first dimension. This minimizes
    the ringing effect of the filter.
    :return: filtered data (mxn numpy array)
    """
    b = b * u.dimensionless_unscaled
    pad_width = 0
    if onset_padding:
        # we do this to minimize ringing of filter
        pad_width = data.shape[0] // 4
        data = np.vstack((data[1:pad_width + 1, :][::-1], data, data[::-1, :][1:pad_width + 1]))
    filtered_data = data
    if not n_jobs:
        n_jobs = np.maximum(joblib.cpu_count(), 1)
    else:
        n_jobs = np.minimum(joblib.cpu_count(), n_jobs)
    if not ch_per_block:
        _virtual_memory = psu.virtual_memory()
        ch_per_block = np.maximum(1, int(_virtual_memory.available /
                                         (filtered_data.nbytes / filtered_data.shape[1]) * 0.1))
    ch_per_block = np.minimum(filtered_data.shape[1], ch_per_block)
    print('filtering {:} channels per block'.format(ch_per_block))

    for _bl in np.arange((np.ceil(filtered_data.shape[1]) / ch_per_block)):
        pos_ini = int(_bl * ch_per_block)
        pos_end = int(np.minimum((_bl + 1) * ch_per_block, filtered_data.shape[1]))
        _b_size = pos_end - pos_ini
        _ch_per_thread = int(np.ceil(float(_b_size) / n_jobs))
        _n_jobs = int(np.minimum(n_jobs, np.ceil(float(_b_size) / _ch_per_thread)))
        # print(('filtering {:} channels per job ({:} parallel jobs)'.format(_ch_per_thread, _n_jobs)))
        sub_pos_ini = []
        sub_pos_end = []
        for _t in range(_n_jobs):
            sub_pos_ini.append(int(pos_ini + _ch_per_thread * _t))
            sub_pos_end.append(int(np.minimum(pos_end, pos_ini + _ch_per_thread * (_t + 1))))

        ParallelTqdm(total=_n_jobs,
                     desc='Parallel filtering',
                     n_jobs=_n_jobs,
                     mmap_mode='w+',
                     backend='threading')(
            joblib.delayed(filt_filt_worker_ovs)(filtered_data, sub_pos_ini[_i], sub_pos_end[_i], b) for
            _i in range(_n_jobs))
        # print('Filtered channels: {:} to {:}'.format(pos_ini, pos_end - 1))
    return filtered_data[pad_width: filtered_data.shape[0] - pad_width, :]



__author__ = 'jundurraga-ucl'
import numpy as np
import peegy.processing.tools.filters.eegFiltering as ep
import matplotlib.pyplot as plt
from astropy import units as u

fs = 44100.0 * u.Hz
t = np.arange(0, 5.5, 1/fs.value) * u.s
y = np.tile(np.array(np.sin(2*np.pi*u.rad * 50 * u.Hz * t)) + np.random.random(t.shape), (4, 1)).T * u.uV

yf = np.fft.fft(y, axis=0)
freq = np.arange(len(yf)) * fs / len(yf)
yc = ep.eeg_notch_filter(x=y, f=[50 * u.Hz, 100 * u.Hz], f_range=5.0 * u.Hz, fs=fs, blocks=8)
plt.plot(t, y)
plt.plot(t, np.squeeze(yc), 'r')
plt.show()
ycf = np.fft.fft(yc, axis=0)
plt.plot(freq, np.abs(yf))
plt.plot(freq, np.abs(ycf), 'r')
plt.show()



# __author__ = 'jundurraga-ucl'
from scipy import signal
import matplotlib.pyplot as plt
import numpy as np
n = 16384 * 1
fs = 16384.0
high_pass = 1.0
low_pass = 100.0

b, a = signal.iirfilter(4, rs=20, Wn=[2.0 * high_pass/fs],
                        ftype='cheby2', btype='highpass')
w, h = signal.freqz(b, a, 8000)
fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(fs / 2 * w / np.pi, 20 * np.log10(abs(h)))
ax.set_xscale('log')
ax.set_title('Chebyshev Type II bandpass frequency response')
ax.set_xlabel('Frequency [radians / second]')
ax.set_ylabel('Amplitude [dB]')
ax.grid(which='both', axis='both')
plt.axhline(y=-3)
plt.axhline(y=-6)
plt.show()



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 16 10:55:37 2014

@author: jundurraga-ucl
"""
from peegy.processing.statistics import eeg_statistic_tools as espt
import numpy as np
import matplotlib.pyplot as plt
# Enable below for interactive backend
# import matplotlib
# if 'qt5agg' in matplotlib.backends.backend_registry.list_builtin():
#     matplotlib.use('qt5agg')

# create synthetic data
fs = 1000.0
nsamples = np.round(1 * fs).astype(int)
nchans = 36
ntrials = 49
noise_dim = 300  # dimensionality of noise
f1 = 40
source = np.expand_dims(np.sin(2 * np.pi * f1 * np.arange(nsamples) / fs), axis=1)
coeff = np.ones(nchans//2) * 0.5 / (nchans / 2)
coeff = np.expand_dims(np.hstack((coeff, coeff)), 0)

s = source * coeff
s_std = np.std(s, axis=0)
s = np.tile(np.expand_dims(s, axis=2), (1, 1, ntrials))

desired_snr = 5.0
ini_std = 10.0 ** (-desired_snr / 20.0) * s_std * ntrials ** 0.5
theoretical_rn = ini_std / ntrials ** 0.5

noise = np.random.normal(0, ini_std[0], size=(nsamples, nchans, ntrials))
# s[:, 0] = s[:, 0] * 0.5
data = noise + s
testg = espt.goertzel(data, fs, frequency=40)
amp, plv, z, z_crit, p_value, angles, dof, rn = espt.phase_locking_value(data, alpha=0.001)
amp_2, plv_2, z_2, z_crit_2, p_value_2, angles_2, dof_2, rn_2 = espt.discrete_phase_locking_value(data,
                                                                                                  alpha=0.001,
                                                                                                  fs=fs,
                                                                                                  frequency=40)
noise_amp, noise_plv, noise_z, noise_z_crit, noise_p_value, noise_angles, noise_dof, noise_rn = \
    espt.phase_locking_value(noise, alpha=0.001)

plt.figure()
plt.plot(plv[40])
plt.plot(plv_2.T)

plt.figure()
plt.plot(angles[40])
plt.plot(angles_2.T)

plt.figure()
plt.plot(amp[40])
plt.plot(amp_2.T)
plt.plot(rn[40])
plt.plot(noise_amp[40])


plt.plot(rn_2.T)
plt.show()

freq = np.arange(0, plv.shape[0]) * fs / data.shape[0]
fig = plt.figure()
ax = fig.add_subplot(141)
ax.plot(source)

ax = fig.add_subplot(142)
ax.plot(freq, plv)

ax = fig.add_subplot(143)
ax.plot(freq, z)
[ax.axhline(_z) for _z in z_crit.squeeze()]

ax = fig.add_subplot(144, projection='polar')
significant = z > z_crit
ax.plot(angles[significant] + np.pi/2, plv[significant], 'o')
ax.plot(angles[40, :] + np.pi/2, plv[40, :], 'o', 'red')
plt.show()



from pathlib import Path
default_path = Path.home().joinpath('peegy')
test_data_path = Path.home().joinpath('peegy').joinpath('test')
test_figures_path = Path.home().joinpath('peegy').joinpath('test').joinpath('figures')



# -*- coding: utf-8 -*-
from peegy.io.exporters.edf_bdf_writer import split_bdf_by_event_code
from peegy.io.external_tools.aep_gui.dataReadingTools import get_files_and_meta_data

import matplotlib
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')

if __name__ == "__main__":
    eeg_data_directory = r'/home/jundurraga/Documents/Measurements/peegy_data/IPM-FR/'
    to_process = get_files_and_meta_data(eeg_data_directory)

    for _, _data_links in to_process.iterrows():
        split_bdf_by_event_code(input_file_name=_data_links.data_links.data_file,
                                event_code=16.0,
                                trash_original=True)



import numpy as np
import peegy.processing.tools.eeg_epoch_operators as eo
from peegy.tools.units.unit_tools import set_default_unit
import astropy.units as u
import warnings
from scipy.linalg import cholesky, eigh

__author__ = 'jundurraga'


def nt_dss0(c0: np.array = np.array([]),
            c1: np.array = np.array([]),
            keep0: int | None = None,
            keep1: float = 1e-9,
            perc0: float = 1,
            perc1: float | None = None
            ):
    """
    Generate DSS matrix
    :param c0: average covariance matrix across all epochs
    :param c1: covariance matrix of average data
    :param keep0: number of components to keep. This parameter is useful to remove negligent components and to add
    numerical stability.
    :param keep1: float indicating the a threshold (relative to the maximum eigenvalue) to keep components and to add
    numerical stability.
    :param perc0: float indicating a percentage of explained variance that would be used as a threshold to keep
    components when whitening data.
    :param perc1: float indicating a percentage of explained variance that would be used as a threshold to keep
    components when generating dss rotation.
    :return: dss matrix, unbiased power, and biased power.
    """
    # make sure covariance matrix is symmetric_positive_definite
    c0 = nearest_symmetric_positive_definite(c0)
    c1 = nearest_symmetric_positive_definite(c1)
    # whitening and selection of components for stability
    if perc0 is not None:
        keep0 = int(c0.shape[0] * perc0)
    topcs_1, evs_1 = nt_pcarot(c0, keep0)
    evs_1 = np.abs(evs_1)
    if keep1 is not None:
        idx = np.where(evs_1 / np.max(evs_1) > keep1)[0]
        topcs_1 = topcs_1[:, idx]
        evs_1 = evs_1[idx]

    # apply PCA and whitening to the biased covariance and ensure that non-negative eigen values are produced
    n1 = np.diag(np.sqrt(1 / evs_1))
    c2 = n1.T.dot(topcs_1.T).dot(c1).dot(topcs_1).dot(n1)
    # matrix to convert PCA-whitened data to DSS
    topcs_2, evs_2 = nt_pcarot(c2, keep0)

    if perc1 is not None:
        n_evs_2 = evs_2 / np.sum(evs_2)
        _thr2 = np.argwhere(np.cumsum(n_evs_2) > perc1)
        _n2 = _thr2[0] if _thr2.size else evs_2.size - 1
        _idx_to_keep2 = np.arange(_n2 + 1)
        topcs_2 = topcs_2[:, _idx_to_keep2]
        # evs_2 = evs_2[_idx_to_keep2]

    # DSS matrix (raw data to normalized DSS)
    todss = topcs_1.dot(n1).dot(topcs_2)
    n2 = np.diagonal(todss.T.dot(c0).dot(todss))
    if np.any(n2 <= 0):
        warnings.warn("""
        Negative or zero normalization value in DSS filter.
        This is usually caused by numerical errors on very low eigenvalue components. Please check the whitening
        threshold. Usually increasing this threshold helps to overcome this issue.
        """)

    todss = todss.dot(np.diag(1 / np.sqrt(n2)))  # adjust so that components are normalized
    # number of components per rotation
    n_0 = evs_1.size
    n_1 = evs_2.size
    # power per DSS component
    pwr0 = np.sqrt(np.sum((c0.T.dot(todss) ** 2), axis=0))  # unbiased
    pwr1 = np.sqrt(np.sum((c1.T.dot(todss) ** 2), axis=0))  # biased
    return todss, pwr0, pwr1, n_0, n_1


def nt_pcarot(cov: np.array = np.array([]),
              n: int = 0):
    eigen_val, eigen_vec = np.linalg.eig(cov)
    eigen_vec = np.real(eigen_vec)
    eigen_val = np.real(eigen_val)
    idx = np.argsort(eigen_val)[::-1]
    if n:
        idx = idx[0: n]
    return eigen_vec[:, idx], eigen_val[idx]


def nt_bias_fft(x: np.array = np.array([]),
                normalized_frequencies: np.array = np.array([]),
                w: np.array = np.array([])):
    if w.size == 0:
        w = np.ones(x.shape)
    c0, _ = eo.et_covariance(x, w=w)
    c1, _ = eo.et_freq_weighted_cov(x, w=w, normalized_frequencies=normalized_frequencies)
    return c0, c1


def nt_bias_fft2(x: np.array = np.array([]),
                 normalized_frequencies: np.array = np.array([]),
                 w: np.array = np.array([])):
    if w.size == 0:
        w = np.ones(x.shape)
    w_ave = eo.w_mean(x, weights=w)
    c0, t0 = eo.et_freq_weighted_cov(x, normalized_frequencies=normalized_frequencies)
    c1, t1 = eo.et_freq_weighted_cov(w_ave, normalized_frequencies=normalized_frequencies)
    return c0, c1, t0, t1


def nearest_symmetric_positive_definite(cov: type(np.array) | None = None):
    """
    Estimate the nearest symmetric positive definite covariance matrix.
    Implementation based on https://se.mathworks.com/matlabcentral/fileexchange/42885-nearestspd?s_tid=mwa_osa_a
    :param cov: the covariance matrix
    :return: the nearest symmetric positive definite covariance matrix.
    """
    cov = set_default_unit(cov, u.dimensionless_unscaled)
    # symmetrize cov into sym_cov
    sym_cov = (cov + cov.T) / 2
    # Compute the symmetric polar factor of sym_cov as h, which is SPD
    _u, _s, _v = np.linalg.svd(sym_cov)
    _h = _v.T.dot(np.diag(_s)).dot(_v)
    # get cov_hat in the above formula
    cov_hat = (sym_cov + _h) / 2

    # ensure symmetry
    cov_hat = (cov_hat + cov_hat.T) / 2
    # test that cov_hat is in fact PD. If it is not so, then tweak it just a bit.
    is_spd = False
    k = 0
    while not is_spd:
        try:
            cholesky(cov_hat)
            is_spd = True
        except np.linalg.LinAlgError:
            is_spd = False
        k = k + 1
        if not is_spd:
            # Ahat failed the chol test. It must have been just a hair off,
            # due to floating point trash, so it is simplest now just to
            # tweak by adding a tiny multiple of an identity matrix.
            eig_val_cov = eigh(cov_hat, eigvals_only=True)
            min_eig = np.min(eig_val_cov)
            cov_hat = cov_hat + (-min_eig * k ** 2 + np.spacing(min_eig)) * np.eye(*cov.shape) * cov.unit
            print('Finding nearest symmetric positive definite matrix iter: {:}'. format(k))
    return cov_hat



from peegy.processing.pipe.definitions import InputOutputProcess
from peegy.processing.tools.eeg_epoch_operators import w_mean
import numpy as np
import os
import astropy.units as u
from typing import List
from PyQt5.QtCore import QLibraryInfo
os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = QLibraryInfo.location(QLibraryInfo.PluginsPath)


class AppendGFPChannel(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 channel_label: str = 'GFP',
                 **kwargs):
        """
        This InputOutputProcess will append the global field power (standard deviation across all channels)
        Lehmann, D., and W. Skrandies. 1980. “Reference-Free Identification of Components of Checkerboard-Evoked
        Multichannel Potential Fields.” Electroencephalography and Clinical Neurophysiology 48 (6): 609–21.
        https://doi.org/10.1016/0013-4694(80)90419-8.
        :param input_process: InputOutputProcess Class
        :param channel_label: name of new channel in layout
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(AppendGFPChannel, self).__init__(input_process=input_process, **kwargs)
        self.channel_label = channel_label

    def transform_data(self):
        data = self.input_node.data
        weights = self.input_node.w
        original_ndim = data.ndim
        # if data is an average rather than single epochs we add an extra dimension
        if original_ndim == 2:
            data = data[:, :, None]
            weights = np.ones(data.shape) * u.dimensionless_unscaled
        if weights is None:
            weights = np.ones(data.shape) * u.dimensionless_unscaled

        average = w_mean(data, weights=weights, keepdims=True, axis=1)
        variance = np.sum(weights * (data - average) ** 2,
                          axis=1, keepdims=True) / np.sum(weights, axis=1, keepdims=True)
        new_channel_data = np.sqrt(variance * data.shape[1])
        self.output_node.data = data
        self.output_node.append_new_channel(new_data=new_channel_data,
                                            layout_label=self.channel_label)
        if original_ndim == 2:
            self.output_node.data = np.squeeze(self.output_node.data)
        if weights is not None:
            new_w = np.ones((weights.shape[0], 1, weights.shape[2])) * weights.unit
            self.output_node.w = np.hstack((weights, new_w))
        self.output_node.statistical_tests = self.input_process.output_node.statistical_tests


class AppendChannels(InputOutputProcess):
    def __init__(self, new_data: type(np.array) | None = None,
                 channel_labels: List[str] | None = None,
                 **kwargs):
        """
        This InputOutputProcess will append the global field power (standard deviation across all channels)
        :param input_process: InputOutputProcess Class
        :param channel_label: name of new channel in layout
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(AppendGFPChannel, self).__init__(input_process=None, **kwargs)
        self.channel_labels = channel_labels
        self.new_data = new_data

    def transform_data(self):
        data = self.input_node.data
        assert data.shape[0] == self.new_data.shape[0]
        assert len(self.channel_labels) == self.new_data.shape[1]
        self.output_node.data = data
        for _data, channel in zip(self.new_data, self.channel_labels):
            self.output_node.append_new_channel(new_data=self.new_data,
                                                layout_label=self.channel_labels)


class CombineChannels(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 new_channel_label: str | None = None,
                 channel_labels: List[str] | None = None,
                 function: object = np.mean,
                 **kwargs):
        """
        This InputOutputProcess will append a new channel based on an operational function passed and applied to a
        subset of existing channels
        :param input_process: InputOutputProcess Class
        :param channel_labels: name of channels that will be used to apply the passed function
        :param new_channel_label: the name of the new channel
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(CombineChannels, self).__init__(input_process=input_process, **kwargs)
        self.channel_labels = channel_labels
        self.new_channel_label = new_channel_label
        self.function = function

    def transform_data(self):
        data = self.input_node.data
        _idx_ch = self.input_node.get_channel_idx_by_label(self.channel_labels)
        _new_channel_label = self.new_channel_label
        if _new_channel_label is None:
            _new_channel_label = '_'.join([_l for _l in self.channel_labels])

        # if data is an average rather than single epochs we add an extra dimension
        weights = self.input_node.w
        original_ndim = data.ndim
        # if data is an average rather than single epochs we add an extra dimension
        if original_ndim == 2:
            data = data[:, :, None]
            weights = np.ones(data.shape) * u.dimensionless_unscaled
        if weights is None:
            weights = np.ones(data.shape) * u.dimensionless_unscaled

        new_channel_data = self.function(data[:, _idx_ch, :])
        self.output_node.data = data
        self.output_node.append_new_channel(new_data=new_channel_data,
                                            layout_label=_new_channel_label)
        if original_ndim == 2:
            self.output_node.data = np.squeeze(self.output_node.data)
        if weights is not None:
            new_w = np.ones((weights.shape[0], 1, weights.shape[2])) * weights.unit
            self.output_node.w = np.hstack((weights, new_w))
        self.output_node.statistical_tests = self.input_process.output_node.statistical_tests



import numpy as np
import peegy.processing.tools.filters.eegFiltering as eegf
import scipy.signal as signal
import time
import matplotlib.pyplot as plt
fs = 44100.0
f1 = 30.0
f2 = 200.0
low_pass = 50.0
high_pass = 100.0
low = 2 * low_pass / fs
high = 2 * high_pass / fs
t = np.arange(0, 7000000) / fs
s = np.sin(2 * np.pi * f1 * t) + np.sin(2 * np.pi * f2 * t)
data = np.tile(s, [24, 1]).T

_b_l, _a_l = signal.butter(N=2, Wn=low, btype='lowpass', analog=False)
filtered_data = eegf.filt_filt_multithread(input_data=data, b=_b_l, a=_a_l)
plt.plot(t, data[:, 0])
plt.plot(t, filtered_data[:, 0])
plt.show()

_b_l, _a_l = signal.butter(N=2, Wn=high, btype='highpass', analog=False)
filtered_data = eegf.filt_filt_multithread(input_data=data, b=_b_l, a=_a_l)
plt.plot(t, data[:, 0])
plt.plot(t, filtered_data[:, 0])
plt.show()

time1 = time.time()
filtered_signal_l2 = signal.filtfilt(_b_l, _a_l, data, axis=0)
time2 = time.time()
print(('standard filter took %f [ms]' % ((time2-time1) * 1000.0)))



import os
import dropbox
import json
import webbrowser
import base64
import requests
import pathlib


def open_url(app_key: str | None = None):
    url = f'https://www.dropbox.com/oauth2/authorize?client_id={app_key}&' \
          f'response_type=code&token_access_type=offline'
    webbrowser.open(url)


def refresh_token(app_key_var_name: str | None = None,
                  app_secret_var_name: str | None = None,
                  access_generated_code_var_name: str | None = None,
                  local_tokens_path: pathlib.Path | None = None,
                  refresh_token_var_name: str | None = None):
    app_key = get_env_variable(env_variable=app_key_var_name, local_tokens_path=local_tokens_path)
    app_secret = get_env_variable(env_variable=app_secret_var_name, local_tokens_path=local_tokens_path)
    access_generated_code = get_env_variable(env_variable=access_generated_code_var_name,
                                             local_tokens_path=local_tokens_path)
    refresh_token = get_env_variable(env_variable=refresh_token_var_name, local_tokens_path=local_tokens_path)
    dbx = None
    response = None
    if refresh_token is not None:
        try:
            dbx = dropbox.Dropbox(
                app_key=app_key,
                app_secret=app_secret,
                oauth2_refresh_token=refresh_token)
        except ValueError:
            pass
    if dbx is None:
        response = renew_access_key(app_key=app_key,
                                    app_secret=app_secret,
                                    access_generated_code=access_generated_code)
        dbx = dropbox.Dropbox(response['refresh_token'])
    return response, dbx


def renew_access_key(app_key: str | None = None,
                     app_secret: str | None = None,
                     access_generated_code: str | None = None
                     ):

    basic_auth = base64.b64encode(f'{app_key}:{app_secret}'.encode())

    headers = {
        'Authorization': f"Basic {basic_auth}",
        'Content-Type': 'application/x-www-form-urlencoded',
    }
    data = f'code={access_generated_code}&grant_type=authorization_code'
    _response = requests.post('https://api.dropboxapi.com/oauth2/token',
                              data=data,
                              auth=(app_key, app_secret),
                              headers=headers)
    response = json.loads(_response.text)

    if 'error' in response.keys():
        open_url(app_key=app_key)
    _response = requests.post('https://api.dropboxapi.com/oauth2/token',
                              data=data,
                              auth=(app_key, app_secret))
    response = json.loads(_response.text)
    return response


def upload_file(input_file: str | None = None,
                remote_file: str | None = None,
                dbx: object | None = None
                ):
    if dbx is None:
        return
    dbx.users_get_current_account()

    for entry in dbx.files_list_folder('').entries:
        print(entry.name)

    with open(input_file):
        _f_bytes = open(input_file, mode="rb")
        data = _f_bytes.read()
        r = dbx.files_upload(data, remote_file, mode=dropbox.files.WriteMode.overwrite)

        try:
            settings = dropbox.sharing.SharedLinkSettings(
                requested_visibility=dropbox.sharing.RequestedVisibility.public,
                audience=dropbox.sharing.LinkAudience.public,
                access=dropbox.sharing.RequestedLinkAccessLevel.viewer,
                allow_download=True
            )
            shared_link_metadata = dbx.sharing_create_shared_link_with_settings(settings=settings,
                                                                                path=r.path_display)
            shared_link = shared_link_metadata.url
        except dropbox.exceptions.ApiError:
            existing_link = dbx.sharing_list_shared_links(path=r.path_display)
            shared_link = existing_link.links[0].url
    shared_file_name = pathlib.Path(input_file).name
    return shared_file_name, shared_link


def get_env_variable(env_variable: str | None = None, local_tokens_path: str | None = None):
    token = os.getenv(env_variable)
    if token is None:
        try:
            print('Could not find {:} in environment. Trying local path {:}'.format(env_variable, local_tokens_path))
            token = json.load(open(local_tokens_path))[env_variable]
        except FileNotFoundError as e:
            print(e.strerror, e.filename)
            token = None
    else:
        print('TOKEN found in $({:})'.format(env_variable))
    return token
